class PreprocessingBase:
    def __init__(self, folder_path):
        self.folder_path = folder_path
        if folder_path is None:
            raise ValueError("folder_path cannot be None")

    def common_preprocessing(self):
        raise NotImplementedError("Must implement common_preprocessing method.")



from sklearn.base import BaseEstimator
from sklearn.base import ClusterMixin


class CustomClusterer(BaseEstimator, ClusterMixin):
    def fit(self, X, y=None):
        raise NotImplementedError("CustomClusterer.fit is not implemented yet.")

    def predict(self, X):
        raise NotImplementedError("CustomClusterer.predict is not implemented yet.")

    def fit_predict(self, X, y=None):
        raise NotImplementedError(
            "CustomClusterer.fit_predict is not implemented yet."
        )



import re

import nltk
import numpy as np
import pandas as pd
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import LabelEncoder

from accelera.src.automl.core.training_tabular_preprocessing_base import (
    TrainingTabularPreprocessingBase,
)
from accelera.src.automl.wrappers.tabular_preprocessing_report import (
    TabularPreprocessingReport,
)
from accelera.src.automl.wrappers.target_classification import TargetClassification
from accelera.src.automl.wrappers.text_graph import TextGraph
from accelera.src.utils.preprocessing import save_pickle


class TextTrainingPreprocessing(TrainingTabularPreprocessingBase):
    def __init__(
        self,
        df,
        target_col: str,
        text_col: str,
        folder_path=None,
        val_size=0.2,
        random_state=42,
        tfidf_max_features=5000,
        tfidf_ngram=(1, 2),
        tfidf_max_df=0.85,
        tfidf_min_df=3,
    ):
        super().__init__(df, target_col, val_size, random_state, folder_path)
        self.text_col = text_col
        self.tfidf_max_features = tfidf_max_features
        self.tfidf_ngram = tfidf_ngram
        self.tfidf_max_df = tfidf_max_df
        self.tfidf_min_df = tfidf_min_df
        if target_col == text_col:
            raise ValueError("target column and text column must not be the same")
        if text_col not in df.columns:
            raise ValueError(
                f"Text column {text_col} not found in dataframe columns"
            )
        if df[text_col].dtype != "object":
            raise ValueError(f"Text column {text_col} must be of type object")
        if not (
            np.issubdtype(self.target_type, np.integer)
            or self.target_type == "object"
            or self.target_type == "bool"
        ):
            raise ValueError(
                "Target must be integer or object fro classification problem"
            )
        self.info = {"target_col": target_col, "text_col": text_col}
        unwanted_columns = []
        for col in self.df.columns:
            if col not in [self.target_col, self.text_col]:
                unwanted_columns.append(col)
        self.df.drop(columns=unwanted_columns, inplace=True)

        nltk.download("punkt")
        nltk.download("stopwords")

    def make_graphs(self, X_train, y_train):
        new_df = X_train.copy()
        new_df[self.target_col] = y_train
        self.report_data["graphs"] = {
            "folder_path": self.folder_path,
            "images_name": [],
        }
        text_graph = TextGraph(
            new_df,
            col_name=self.text_col,
            target_name=self.target_col,
            folder_path=self.folder_path,
        )
        text_graph.build_graph()
        self.report_data["graphs"]["images_name"].append(self.text_col)

        target_graph = TargetClassification(
            new_df,
            col_name=self.target_col,
            target_name=self.target_col,
            folder_path=self.folder_path,
        )
        target_graph.build_graph()
        self.report_data["graphs"]["images_name"].append(self.target_col)

    def custom_text_tokenizer(self, text):
        stop_words = set(stopwords.words("english")) - {"not", "no"}
        stem = PorterStemmer()
        text = text.lower().strip()
        text = re.sub(r"’", "'", text)
        text = re.sub(r"can[o']t", "can not", text)
        text = re.sub(r"won[o']t", "will not", text)
        text = re.sub(r"shan[o']t", "shall not", text)
        text = re.sub(r"n't\b", " not ", text)
        text = re.sub(r"(\d)([a-zA-Z])", r"\1 \2", text)
        text = re.sub(r"([a-zA-Z])(\d)", r"\1 \2", text)
        text = re.sub(r"[^a-zA-Z0-9']", " ", text)
        text = re.sub(r"\s+", " ", text)
        tokens = [word.strip("'").strip('"').strip() for word in text.split()]
        tokens = [
            stem.stem(word)
            for word in tokens
            if word not in stop_words and len(word) > 1
        ]
        return tokens

    def features_preprocessing(self, X_train, X_val):
        X_train[self.text_col] = X_train[self.text_col].fillna("")
        X_val[self.text_col] = X_val[self.text_col].fillna("")
        self.report_data["preprocessing"].append(
            {
                "col_name": self.text_col,
                "col_type": "Text",
                "col_preprocessing": [
                    "Fill nulls with empty text",
                    "Cleaning & Tokeniztion & Stemming",
                    "Tfidf Vectorizer",
                ],
            }
        )
        X_train_head = (
            X_train[self.text_col]
            .head()
            .reset_index(drop=True)
            .to_frame(name=self.text_col)
        )
        X_val_head = (
            X_val[self.text_col]
            .head()
            .reset_index(drop=True)
            .to_frame(name=self.text_col)
        )
        X_train_head_cleaned = (
            X_train_head[self.text_col]
            .apply(lambda x: " ".join(self.custom_text_tokenizer(x)))
            .to_frame(
                name=f"{self.text_col} After Cleaning & Tokeniztion & Stemming"
            )
        )
        X_val_head_cleaned = (
            X_val_head[self.text_col]
            .apply(lambda x: " ".join(self.custom_text_tokenizer(x)))
            .to_frame(
                name=f"{self.text_col} After Cleaning & Tokeniztion & Stemming"
            )
        )
        self.report_data["after_preprocessing"] = {
            "X_train_processed": pd.concat(
                [X_train_head, X_train_head_cleaned], axis=1
            ),
            "X_val_processed": pd.concat([X_val_head, X_val_head_cleaned], axis=1),
        }
        training_preprocessor = TfidfVectorizer(
            max_features=self.tfidf_max_features,
            ngram_range=self.tfidf_ngram,
            max_df=self.tfidf_max_df,
            min_df=self.tfidf_min_df,
            tokenizer=self.custom_text_tokenizer,
            token_pattern=None,
            lowercase=False,
        )

        X_train_preprocessed = training_preprocessor.fit_transform(
            X_train[self.text_col]
        )
        X_val_preprocessed = training_preprocessor.transform(X_val[self.text_col])

        save_pickle(
            self.folder_path, training_preprocessor, "training_preprocessor.pkl"
        )
        return X_train_preprocessed, X_val_preprocessed

    def target_preprocessing(self, y_train, y_val):
        self.report_data["preprocessing"].append(
            {
                "col_name": self.target_col,
                "col_type": "Classification",
                "col_preprocessing": [
                    "Fill missing with most frequent",
                    "Label encoding",
                ],
            }
        )
        mode = y_train.mode()[0]
        self.info["target_mode"] = mode
        y_train = y_train.fillna(mode)
        y_val = y_val.fillna(mode)
        label_encoder = LabelEncoder()
        y_train_preprocessed = label_encoder.fit_transform(y_train)
        y_val_preprocessed = label_encoder.transform(y_val)
        self.report_data["after_preprocessing"]["y_train_processed"] = pd.DataFrame(
            y_train_preprocessed, columns=[self.target_col]
        ).head()
        self.report_data["after_preprocessing"]["y_val_processed"] = pd.DataFrame(
            y_val_preprocessed, columns=[self.target_col]
        ).head()
        save_pickle(self.folder_path, label_encoder, "target_preprocessor.pkl")
        save_pickle(self.folder_path, self.info, "data_info.pkl")
        return y_train_preprocessed, y_val_preprocessed

    def common_preprocessing(self):
        self.report_data["drop_columns"] = {"col_drop": None}
        self.data_overview()
        self.drop_duplicates()
        X_train, X_val, y_train, y_val = self.split_data()
        self.make_graphs(X_train, y_train)
        self.report_data["preprocessing"] = []
        X_train, X_val = self.features_preprocessing(X_train, X_val)
        y_train, y_val = self.target_preprocessing(y_train, y_val)
        report = TabularPreprocessingReport(
            self.folder_path, self.report_data, text_based=True
        )
        report.execute()
        return X_train, y_train, X_val, y_val



import pandas as pd

from accelera.src.accelera_pipe.core.metric_display import MetricDisplay


class DisplayArraySingle(MetricDisplay):
    def __init__(self, metric_name, values):
        super().__init__(metric_name, values)

    def execute(self):
        content = f"<div>\n<h3>Metric name: {self.metric_name}</h3>\n"
        ids = [value["metric id"] for value in self.values]
        labels_name = self.handle_name("labels_name", len(self.values[0]["result"]))
        data = {"Metric ID": ids}
        for value in self.values:
            for i in range(len(value["result"])):
                if labels_name[i] not in data:
                    data[labels_name[i]] = [value["result"][i]]
                else:
                    data[labels_name[i]].append(value["result"][i])

        table = pd.DataFrame(data).to_html(index=False, border=1, justify="center")
        content = content + table + "</div>\n"
        return content



import random

import numpy as np
import torch
from PIL import Image

from accelera.src.automl.core.image_dataset import ImageDataset


class SegmentationImageDataset(ImageDataset):
    def __init__(
        self,
        image_paths,
        masks=None,
        image_size=(224, 224),
        binary_mask_threshold=128,
        augment=True,
        augmentation_probability=0.5,
        horizontal_flip=True,
        vertical_flip=True,
        rotation=True,
        rotation_angle=30,
        brightness=True,
        brightness_factors=(0.8, 1.2),
        contrast=True,
        contrast_factors=(0.8, 1.2),
    ):
        super().__init__(
            image_paths,
            masks,
            image_size,
            augment,
            augmentation_probability,
            horizontal_flip,
            vertical_flip,
            rotation,
            rotation_angle,
            brightness,
            brightness_factors,
            contrast,
            contrast_factors,
        )
        self.binary_mask_threshold = binary_mask_threshold

    def load_image_masks(self, index):
        path = self.image_paths[index]
        img = Image.open(path).convert("RGB")
        img = img.resize(self.image_size, resample=Image.BILINEAR)
        mask = None
        if self.labels is not None:
            mask_path = self.labels[index]
            mask = Image.open(mask_path).convert("L")
            mask = mask.resize(self.image_size, resample=Image.NEAREST)

        if self.augment:
            img, mask = self.augmentation(img, mask)
        img_array = np.array(img, dtype=np.float32) / 255.0
        img_transpose = np.transpose(img_array, (2, 0, 1))
        img_tensor = torch.tensor(img_transpose, dtype=torch.float32)
        mask_tensor = None
        if mask is not None:
            mask_array = np.array(mask, dtype=np.float32)
            mask_array = (mask_array >= self.binary_mask_threshold).astype(
                np.float32
            )
            mask_tensor = torch.tensor(mask_array, dtype=torch.float32).unsqueeze(0)
        return img_tensor, mask_tensor

    def random_horizontal_flip(self, img, mask):
        transposed_img = img
        transposed_mask = mask
        if self.horizontal_flip and random.random() < self.augmentation_probability:
            transposed_img = img.transpose(Image.FLIP_LEFT_RIGHT)
            if mask is not None:
                transposed_mask = mask.transpose(Image.FLIP_LEFT_RIGHT)
        return transposed_img, transposed_mask

    def random_vertical_flip(self, img, mask):
        transposed_img = img
        transposed_mask = mask
        if self.vertical_flip and random.random() < self.augmentation_probability:
            transposed_img = img.transpose(Image.FLIP_TOP_BOTTOM)
            if mask is not None:
                transposed_mask = mask.transpose(Image.FLIP_TOP_BOTTOM)
        return transposed_img, transposed_mask

    def random_rotation(self, img, mask):
        rotated_img = img
        rotated_mask = mask
        if self.rotation and random.random() < self.augmentation_probability:
            random_angle = random.uniform(
                -1 * self.rotation_angle, self.rotation_angle
            )
            rotated_img = img.rotate(random_angle)
            if mask is not None:
                rotated_mask = mask.rotate(random_angle, resample=Image.NEAREST)
        return rotated_img, rotated_mask

    def augmentation(self, img, mask):
        img, mask = self.random_horizontal_flip(img, mask)
        img, mask = self.random_vertical_flip(img, mask)
        img, mask = self.random_rotation(img, mask)
        img = self.random_brightness(img)
        img = self.random_contrast(img)
        return img, mask

    def __getitem__(self, index):
        img_tensor, mask_tensor = self.load_image_masks(index)
        return img_tensor, mask_tensor



from accelera.src.automl.wrappers.graph_base import GraphBase


class TabularGraphBase(GraphBase):
    def __init__(self, df, col_name, target_name, folder_path):
        super().__init__(folder_path)
        self.df = df
        self.col_name = col_name
        self.target_name = target_name
        if self.col_name is not None:
            self.nulls_percent = 100 * self.df[col_name].isna().mean()
            self.graph_df = self.df.dropna(subset=[self.col_name])



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src.automl.core import (
    classical_testing_preprocessing as classical_testing_preprocessing,
)
from accelera.src.automl.core import (
    classical_testing_preprocessing_test as classical_testing_preprocessing_test,
)
from accelera.src.automl.core import (
    classical_training_preprocessing as classical_training_preprocessing,
)
from accelera.src.automl.core import (
    classical_training_preprocessing_test as classical_training_preprocessing_test,
)
from accelera.src.automl.core import (
    classification_image_dataset as classification_image_dataset,
)
from accelera.src.automl.core import (
    classification_image_dataset_test as classification_image_dataset_test,
)
from accelera.src.automl.core import (
    classification_image_testing_preprocessing as classification_image_testing_preprocessing,
)
from accelera.src.automl.core import (
    classification_image_testing_preprocessing_test as classification_image_testing_preprocessing_test,
)
from accelera.src.automl.core import (
    classification_image_training_preprocessing as classification_image_training_preprocessing,
)
from accelera.src.automl.core import (
    classification_image_training_preprocessing_test as classification_image_training_preprocessing_test,
)
from accelera.src.automl.core import image_dataset as image_dataset
from accelera.src.automl.core import image_dataset_test as image_dataset_test
from accelera.src.automl.core import (
    image_training_preprocessing as image_training_preprocessing,
)
from accelera.src.automl.core import (
    image_training_preprocessing_test as image_training_preprocessing_test,
)
from accelera.src.automl.core import preprocessing_base as preprocessing_base
from accelera.src.automl.core import (
    segmentation_image_dataset as segmentation_image_dataset,
)
from accelera.src.automl.core import (
    segmentation_image_dataset_test as segmentation_image_dataset_test,
)
from accelera.src.automl.core import (
    segmentation_image_testing_preprocessing as segmentation_image_testing_preprocessing,
)
from accelera.src.automl.core import (
    segmentation_image_testing_preprocessing_test as segmentation_image_testing_preprocessing_test,
)
from accelera.src.automl.core import (
    segmentation_image_training_preprocessing as segmentation_image_training_preprocessing,
)
from accelera.src.automl.core import (
    segmentation_image_training_preprocessing_test as segmentation_image_training_preprocessing_test,
)
from accelera.src.automl.core import (
    tabular_preprocessing_base as tabular_preprocessing_base,
)
from accelera.src.automl.core import (
    testing_tabular_preprocessing_base as testing_tabular_preprocessing_base,
)
from accelera.src.automl.core import (
    text_testing_preprocesing as text_testing_preprocesing,
)
from accelera.src.automl.core import (
    text_testing_preprocessing_test as text_testing_preprocessing_test,
)
from accelera.src.automl.core import (
    text_training_preprocessing as text_training_preprocessing,
)
from accelera.src.automl.core import (
    text_training_preprocessing_test as text_training_preprocessing_test,
)
from accelera.src.automl.core import (
    training_tabular_preprocessing_base as training_tabular_preprocessing_base,
)



import numpy as np


def convert_to_array(x, dtype=None):
    if x is None:
        return None
    if dtype is not None:
        return np.asarray(x, dtype=dtype)
    return np.asarray(x)


def validate_array_shape(array, expected_shape, names=None):
    if names is None:
        names = ["array", "reference"]

    if array.shape[0] != expected_shape:
        raise ValueError(
            f"{names[0]} and {names[1]} must have the same number of samples. "
            f"Got {names[0]} with {array.shape[0]} samples and "
            f"{names[1]} with {expected_shape} samples."
        )


def get_array_info(x):
    if x is None:
        return {"shape": None, "dtype": None, "ndim": None, "size": None}

    x = convert_to_array(x)
    return {"shape": x.shape, "dtype": x.dtype, "ndim": x.ndim, "size": x.size}



import ast


class _CppEmitError(ValueError):
    pass


class _PythonToCpp(ast.NodeVisitor):
    def __init__(self):
        self._lines: list[str] = []
        self._indent = 0
        self._declared: set[str] = set()
        self._includes: set[str] = {"#include <iostream>"}

    def emit(self, line: str = "") -> None:
        self._lines.append(("    " * self._indent) + line)

    def render(self) -> str:
        includes = sorted(self._includes)
        return (
            "\n".join(includes + ([""] if includes else []) + self._lines).rstrip()
            + "\n"
        )

    def expr(self, node: ast.AST) -> str:
        if isinstance(node, ast.Constant):
            if node.value is None:
                return "nullptr"
            if isinstance(node.value, bool):
                return "true" if node.value else "false"
            if isinstance(node.value, str):
                return (
                    '"' + node.value.replace("\\", "\\\\").replace('"', '\\"') + '"'
                )
            return repr(node.value)

        if isinstance(node, ast.Name):
            return node.id

        if isinstance(node, ast.BinOp):
            op = self._binop(node.op)
            return f"({self.expr(node.left)} {op} {self.expr(node.right)})"

        if isinstance(node, ast.UnaryOp):
            op = self._unaryop(node.op)
            return f"({op}{self.expr(node.operand)})"

        if isinstance(node, ast.Compare):
            if len(node.ops) != 1 or len(node.comparators) != 1:
                raise _CppEmitError("chained comparisons are not supported")
            op = self._cmpop(node.ops[0])
            return f"({self.expr(node.left)} {op} {self.expr(node.comparators[0])})"

        if isinstance(node, ast.BoolOp):
            op = "&&" if isinstance(node.op, ast.And) else "||"
            parts = [self.expr(v) for v in node.values]
            return "(" + f" {op} ".join(parts) + ")"

        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name) and node.func.id == "print":
                return self._emit_print_expr(node)

            func = self.expr(node.func)
            args = ", ".join(self.expr(a) for a in node.args)
            return f"{func}({args})"

        if isinstance(node, ast.Attribute):
            return f"{self.expr(node.value)}.{node.attr}"

        if isinstance(node, ast.Subscript):
            return f"{self.expr(node.value)}[{self.expr(node.slice)}]"

        if isinstance(node, ast.Slice):
            raise _CppEmitError("slice syntax is not supported")

        raise _CppEmitError(f"unsupported expression: {type(node).__name__}")

    def _emit_print_expr(self, node: ast.Call) -> str:
        if node.keywords:
            raise _CppEmitError("print() keywords are not supported")
        self._includes.add("#include <iostream>")
        if not node.args:
            return "(std::cout << std::endl)"
        stream = "std::cout"
        for i, arg in enumerate(node.args):
            stream += f" << {self.expr(arg)}"
            if i != len(node.args) - 1:
                stream += ' << " "'
        stream += " << std::endl"
        return f"({stream})"

    def _binop(self, op: ast.operator) -> str:
        table = {
            ast.Add: "+",
            ast.Sub: "-",
            ast.Mult: "*",
            ast.Div: "/",
            ast.Mod: "%",
            ast.Pow: None,
        }
        out = table.get(type(op))
        if out is None:
            raise _CppEmitError(f"unsupported binary operator: {type(op).__name__}")
        return out

    def _unaryop(self, op: ast.unaryop) -> str:
        table = {ast.UAdd: "+", ast.USub: "-", ast.Not: "!"}
        out = table.get(type(op))
        if out is None:
            raise _CppEmitError(f"unsupported unary operator: {type(op).__name__}")
        return out

    def _cmpop(self, op: ast.cmpop) -> str:
        table = {
            ast.Eq: "==",
            ast.NotEq: "!=",
            ast.Lt: "<",
            ast.LtE: "<=",
            ast.Gt: ">",
            ast.GtE: ">=",
        }
        out = table.get(type(op))
        if out is None:
            raise _CppEmitError(
                f"unsupported comparison operator: {type(op).__name__}"
            )
        return out

    def visit_Module(self, node: ast.Module) -> None:
        function_defs = [
            stmt for stmt in node.body if isinstance(stmt, ast.FunctionDef)
        ]
        main_body = [
            stmt for stmt in node.body if not isinstance(stmt, ast.FunctionDef)
        ]

        for stmt in function_defs:
            self.visit(stmt)

        self.emit("int main() {")
        self._indent += 1
        prev_declared = self._declared
        self._declared = set()
        for stmt in main_body:
            self.visit(stmt)
        self.emit("return 0;")
        self._declared = prev_declared
        self._indent -= 1
        self.emit("}")

    def visit_Expr(self, node: ast.Expr) -> None:
        self.emit(f"{self.expr(node.value)};")

    def visit_Assign(self, node: ast.Assign) -> None:
        if len(node.targets) != 1:
            raise _CppEmitError("multi-target assignment is not supported")
        target = node.targets[0]
        if not isinstance(target, ast.Name):
            raise _CppEmitError("only simple name assignment is supported")
        name = target.id
        value = self.expr(node.value)
        if name not in self._declared:
            ctype = self._infer_ctype(node.value)
            self._declared.add(name)
            self.emit(f"{ctype} {name} = {value};")
        else:
            self.emit(f"{name} = {value};")

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        if not isinstance(node.target, ast.Name):
            raise _CppEmitError("only simple name aug-assign is supported")
        name = node.target.id
        op = self._binop(node.op)
        self.emit(f"{name} {op}= {self.expr(node.value)};")

    def visit_Return(self, node: ast.Return) -> None:
        if node.value is None:
            self.emit("return;")
        else:
            self.emit(f"return {self.expr(node.value)};")

    def visit_If(self, node: ast.If) -> None:
        self.emit(f"if {self.expr(node.test)} {{")
        self._indent += 1
        for s in node.body:
            self.visit(s)
        self._indent -= 1
        if node.orelse:
            self.emit("} else {")
            self._indent += 1
            for s in node.orelse:
                self.visit(s)
            self._indent -= 1
        self.emit("}")

    def visit_For(self, node: ast.For) -> None:
        if not isinstance(node.target, ast.Name):
            raise _CppEmitError("only simple for-loop targets are supported")
        var = node.target.id
        start, stop, step = self._parse_range_iter(node.iter)
        if var not in self._declared:
            self._declared.add(var)
        cmp = "<" if step is None or not step.strip().startswith("-") else ">"
        step_expr = step if step is not None else "1"
        self.emit(
            f"for (int {var} = {start}; {var} {cmp} {stop}; {var} += {step_expr}) {{"
        )
        self._indent += 1
        for s in node.body:
            self.visit(s)
        self._indent -= 1
        self.emit("}")

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        if node.decorator_list:
            raise _CppEmitError("decorators are not supported")
        if node.returns is not None:
            pass
        args = []
        for a in node.args.args:
            args.append(f"auto {a.arg}")
        self.emit(f"auto {node.name}({', '.join(args)}) {{")
        self._indent += 1
        prev_declared = self._declared
        self._declared = set(a.arg for a in node.args.args)
        for s in node.body:
            self.visit(s)
        self._declared = prev_declared
        self._indent -= 1
        self.emit("}")
        self.emit("")

    def generic_visit(self, node: ast.AST) -> None:
        raise _CppEmitError(f"unsupported syntax: {type(node).__name__}")

    def _infer_ctype(self, node: ast.AST) -> str:
        if isinstance(node, ast.Constant):
            if isinstance(node.value, bool):
                return "bool"
            if isinstance(node.value, int):
                return "int"
            if isinstance(node.value, float):
                return "double"
            if isinstance(node.value, str):
                self._includes.add("#include <string>")
                return "std::string"
        return "auto"

    def _parse_range_iter(self, node: ast.AST) -> tuple[str, str, str | None]:
        if not (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "range"
            and not node.keywords
        ):
            raise _CppEmitError("for-loop only supports range(...)")
        if not (1 <= len(node.args) <= 3):
            raise _CppEmitError("range() with 1..3 args is supported")

        if len(node.args) == 1:
            return "0", self.expr(node.args[0]), None
        if len(node.args) == 2:
            return self.expr(node.args[0]), self.expr(node.args[1]), None
        return (
            self.expr(node.args[0]),
            self.expr(node.args[1]),
            self.expr(node.args[2]),
        )


def py2cpp_converter(python_code: str) -> str:
    try:
        tree = ast.parse(python_code)
    except SyntaxError as e:
        raise ValueError(f"Invalid python code: {e}") from e

    converter = _PythonToCpp()
    try:
        converter.visit(tree)
    except _CppEmitError as e:
        raise ValueError(str(e)) from e
    return converter.render()



import io
import os

from sklearn.model_selection import train_test_split

from accelera.src.automl.core.tabular_preprocessing_base import (
    TabularPreprocessingBase,
)
from accelera.src.utils.preprocessing import lower_data


class TrainingTabularPreprocessingBase(TabularPreprocessingBase):
    def __init__(self, df, target_col, val_size, random_state, folder_path=None):
        super().__init__(df, folder_path)
        self.target_col = target_col
        self.val_size = val_size
        self.random_state = random_state
        self.report_data = {}

        if self.target_col not in self.df.columns:
            raise ValueError("target_col must be one of the dataframe columns")

        if (not (isinstance(self.val_size, (int, float)))) or (
            not (0 <= self.val_size <= 0.5)
        ):
            raise ValueError(
                "test size is invalid it must be less than or equal 0.5"
            )

        if self.random_state is not None and not (
            isinstance(self.random_state, int)
        ):
            raise ValueError("random state is invalid it must be integer or None")

        self.target_type = self.df[self.target_col].dtype
        os.makedirs(self.folder_path, exist_ok=True)

    def data_overview(self):
        data_head = self.df.head()
        lower_data(self.df)
        lower_data_head = self.df.head()
        io_buffer = io.StringIO()
        self.df.info(buf=io_buffer)
        data_info = io_buffer.getvalue()
        numerical_df = self.df.select_dtypes(include="number")
        categorical_df = self.df.select_dtypes(include="object")
        numerical_describe, categorical_describe = None, None
        if not numerical_df.empty:
            numerical_describe = numerical_df.describe()
        if not categorical_df.empty:
            categorical_describe = categorical_df.describe()
        missing_values = self.df.isnull().sum()
        duplicates_sum = self.df.duplicated().sum()
        duplicates_percentage = self.df.duplicated().mean() * 100
        self.report_data["data_overview"] = {
            "data_head": data_head,
            "lower_data_head": lower_data_head,
            "info": data_info,
            "numerical_describe": numerical_describe,
            "categorical_describe": categorical_describe,
            "missing_values": missing_values,
            "duplicates_sum": duplicates_sum,
            "duplicates_percentage": duplicates_percentage,
            "shape": self.df.shape,
        }

    def drop_duplicates(self):
        self.df.drop_duplicates(inplace=True)
        self.report_data["drop_duplicates"] = {
            "shape": self.df.shape,
            "duplicates_sum": self.df.duplicated().sum(),
            "duplicates_percentage": self.df.duplicated().mean() * 100,
        }

    def split_data(self):
        X, y = self.df.drop(columns=[self.target_col]), self.df[self.target_col]
        X_train, X_val, y_train, y_val = train_test_split(
            X, y, test_size=self.val_size, random_state=self.random_state
        )
        self.report_data["split"] = {
            "val_size": self.val_size,
            "X_train_shape": X_train.shape,
            "X_val_shape": X_val.shape,
            "y_train_shape": y_train.shape,
            "y_val_shape": y_val.shape,
        }
        return X_train, X_val, y_train, y_val



import json
import os
import re
from pathlib import Path

import numpy as np
import requests

from accelera.src.config import config
from accelera.src.utils.py2cpp_converter import py2cpp_converter

try:
    from code_parallelizer_utils import extract_loops as _extract_loops
    from code_parallelizer_utils import write_loops_to_json as _write_loops_to_json
except ImportError:
    _extract_loops = None
    _write_loops_to_json = None


def extract_loops(code: str, clang_args: list | None = None) -> list:
    if clang_args is None:
        clang_args = list(config.DEFAULT_CLANG_ARGS)
    return _extract_loops(code, clang_args)


def write_loops_to_json(loops: list, output_json: str) -> bool:
    return _write_loops_to_json(loops, output_json)


def pragma_to_class(label: str, pragma: str) -> str:
    if label == "False":
        return "none"
    if re.search(r"reduction\s*\(", pragma):
        return "reduction"
    return "parallel_for"


def extract_features(code: str) -> dict:
    features = {
        "has_reduction_plus": False,
        "has_reduction_mul": False,
        "has_reduction_general": False,
        "reduction_var": None,
        "has_loop_carried_dep": False,
        "has_pointer_aliasing": False,
        "control_flow_inside": False,
        "array_writes": [],
        "array_reads": [],
        "loop_bound_constant": False,
        "has_consecutive_nested_loops": False,
        "max_consecutive_loop_depth": 1,
        "has_nested_braces": False,
        "has_raw_dependency": False,
        "has_war_dependency": False,
        "dependency_distance": 0,
        "has_indirect_access": False,
        "stride_pattern": 1,
        "memory_complexity": 0,
        "branch_count": 0,
        "has_early_exit": False,
        "function_call_count": 0,
        "arithmetic_op_count": 0,
        "memory_op_count": 0,
        "is_reduction_max": False,
        "is_reduction_min": False,
        "reduction_var_count": 0,
        "trip_count_computable": False,
        "estimated_iterations": 0,
        "vectorizable": False,
        "read_var_count": 0,
        "write_var_count": 0,
    }

    code_clean = re.sub(
        r"//.*?$|/\*.*?\*/", "", code, flags=re.MULTILINE | re.DOTALL
    )

    if re.search(r"\bfor\b\s*\(.*?\)\s*[{]?\s*\bfor\b", code_clean, re.DOTALL):
        features["has_consecutive_nested_loops"] = True

    for_positions = [m.start() for m in re.finditer(r"\bfor\b", code_clean)]
    max_consecutive = 1
    current = 1
    for i in range(1, len(for_positions)):
        if for_positions[i] - for_positions[i - 1] < 200:
            current += 1
            max_consecutive = max(max_consecutive, current)
        else:
            current = 1
    features["max_consecutive_loop_depth"] = max_consecutive

    brace_depth = 0
    max_brace_depth = 0
    for char in code_clean:
        if char == "{":
            brace_depth += 1
            max_brace_depth = max(max_brace_depth, brace_depth)
        elif char == "}":
            brace_depth -= 1
    features["has_nested_braces"] = max_brace_depth >= 3

    lines = [
        line.strip()
        for line in code_clean.split("\n")
        if "=" in line and line.strip()
    ]
    reduction_vars = set()
    for line in lines:
        if match := re.search(r"(\w+)\s*\+=\s*", line):
            var = match.group(1)
            if not re.search(rf"\b{var}\s*\[", code_clean):
                features["has_reduction_plus"] = True
                features["has_reduction_general"] = True
                features["reduction_var"] = var
                reduction_vars.add(var)
        elif match := re.search(r"(\w+)\s*=\s*\1\s*\+", line):
            var = match.group(1)
            if not re.search(rf"\b{var}\s*\[", code_clean):
                features["has_reduction_plus"] = True
                features["has_reduction_general"] = True
                features["reduction_var"] = var
                reduction_vars.add(var)
        elif match := re.search(r"(\w+)\s*\*=\s*", line):
            var = match.group(1)
            if not re.search(rf"\b{var}\s*\[", code_clean):
                features["has_reduction_mul"] = True
                features["has_reduction_general"] = True
                reduction_vars.add(var)

    features["reduction_var_count"] = len(reduction_vars)

    if re.search(r"\w+\[\s*\w+\s*-\s*1\s*\]", code_clean):
        features["has_loop_carried_dep"] = True
        features["has_raw_dependency"] = True
        features["dependency_distance"] = 1

    writes = re.findall(r"(\w+)\s*\[[^\]]+\]\s*=", code_clean)
    reads = re.findall(r"(\w+)\s*\[[^\]]+\]", code_clean)
    features["array_writes"] = list(set(writes))
    features["array_reads"] = list(set(reads) - set(writes))
    features["write_var_count"] = len(set(writes))
    features["read_var_count"] = len(set(reads))

    if re.search(r"\w+\s*\[\s*\w+\s*\[[^\]]+\]\s*\]", code_clean):
        features["has_indirect_access"] = True
        features["memory_complexity"] = 2

    if matches := re.findall(r"\w+\s*<\s*(\d+)", code_clean):
        if any(int(m) < 100 for m in matches):
            features["loop_bound_constant"] = True
            features["trip_count_computable"] = True
        features["estimated_iterations"] = max([int(m) for m in matches], default=0)

    features["branch_count"] = len(
        re.findall(r"\bif\b|\belse\b|\bswitch\b", code_clean)
    )
    features["control_flow_inside"] = features["branch_count"] > 0

    if re.search(r"\bbreak\b|\bcontinue\b|\breturn\b", code_clean):
        features["has_early_exit"] = True

    features["function_call_count"] = len(re.findall(r"\w+\s*\(", code_clean)) - len(
        re.findall(r"\bfor\b|\bif\b|\bwhile\b", code_clean)
    )

    features["arithmetic_op_count"] = len(
        re.findall(r"[+\-](?!=)", code_clean)
    ) + len(re.findall(r"[*/%](?!=)", code_clean))

    features["memory_op_count"] = len(re.findall(r"\w+\s*\[", code_clean))

    if re.search(r"\bmax\b|\bfmax\b", code_clean):
        features["is_reduction_max"] = True
        features["has_reduction_general"] = True
    if re.search(r"\bmin\b|\bfmin\b", code_clean):
        features["is_reduction_min"] = True
        features["has_reduction_general"] = True

    common_vars = set(reads) & set(writes)
    if common_vars:
        features["has_war_dependency"] = True

    features["vectorizable"] = (
        features["branch_count"] == 0
        and features["function_call_count"] == 0
        and not features["has_loop_carried_dep"]
        and not features["has_indirect_access"]
    )

    return features


def vectorize_features(features: dict) -> np.ndarray:
    vec = np.zeros(40, dtype=np.float32)
    bool_keys = [
        "has_reduction_plus",
        "has_reduction_mul",
        "has_reduction_general",
        "has_loop_carried_dep",
        "has_pointer_aliasing",
        "control_flow_inside",
        "loop_bound_constant",
        "has_consecutive_nested_loops",
        "has_nested_braces",
        "has_raw_dependency",
        "has_war_dependency",
        "has_indirect_access",
        "has_early_exit",
        "is_reduction_max",
        "is_reduction_min",
        "trip_count_computable",
        "vectorizable",
    ]
    for i, key in enumerate(bool_keys):
        vec[i] = 1.0 if features.get(key, False) else 0.0

    vec[17] = min(len(features.get("array_writes", [])), 5) / 5.0
    vec[18] = min(len(features.get("array_reads", [])), 5) / 5.0
    vec[19] = min(features.get("max_consecutive_loop_depth", 1), 4) / 4.0
    vec[20] = min(features.get("max_consecutive_loop_depth", 1), 4) / 4.0

    def hash_array_names(names, dim=5):
        h = np.zeros(dim)
        for name in names[:5]:
            seed = sum(ord(c) for c in name) % 1000
            for j in range(dim):
                h[j] = (h[j] + ((seed * (j + 1)) % 101) / 101.0) % 1.0
        return h / max(1, len(names)) if names else h

    vec[21:26] = hash_array_names(features.get("array_writes", []))
    vec[26:31] = hash_array_names(features.get("array_reads", []))
    vec[31] = 1.0 if features.get("reduction_var") else 0.0
    vec[32] = min(features.get("branch_count", 0), 10) / 10.0
    vec[33] = min(features.get("function_call_count", 0), 10) / 10.0
    vec[34] = min(features.get("arithmetic_op_count", 0), 50) / 50.0
    vec[35] = min(features.get("memory_op_count", 0), 50) / 50.0

    mem_ops = features.get("memory_op_count", 0)
    arith_ops = features.get("arithmetic_op_count", 0)
    if mem_ops > 0:
        ratio = arith_ops / mem_ops
        vec[36] = min(ratio, 5.0) / 5.0
    else:
        vec[36] = 0.0

    vec[37] = min(features.get("read_var_count", 0), 20) / 20.0
    vec[38] = min(features.get("write_var_count", 0), 20) / 20.0
    vec[39] = min(features.get("reduction_var_count", 0), 5) / 5.0

    return vec


def validate_pragma(pragma: str) -> str:
    if not pragma.startswith("#pragma"):
        pragma = f"#pragma {pragma}"

    brackets = {"(": ")", "{": "}", "[": "]"}
    stack = []

    for char in pragma:
        if char in brackets:
            stack.append(char)
        elif char in brackets.values():
            if not stack:
                return pragma.rstrip(char)

            top = stack.pop()
            if brackets[top] != char:
                return pragma.rstrip(char)

    while stack:
        pragma += brackets[stack.pop()]

    return pragma


class Parallelizer:
    def __init__(self):
        self.root_path = config.REPO_ROOT
        self.cache_dir = config.cache_dir
        self.classifier_endpoint = config.CLASSIFIER_ENDPOINT
        self.generator_endpoint = config.GENERATOR_ENDPOINT

    def _classify(self, embedding):
        try:
            response = requests.post(
                self.classifier_endpoint, json={"embedding": embedding.tolist()}
            )
            result = response.json()
            if response.status_code != 200:
                raise RuntimeError(
                    "Error while parallelizing, "
                    "in classifier with error: "
                    f"{result.get('error', 'Unknown error')}"
                )

            return result["result"]
        except Exception as e:
            raise RuntimeError(
                f"Error while parallelizing, in classifier with error: {e}"
            )

    def _generate_omp_pragma_with_loop(self, loop_code: str, loop_class: str) -> str:
        if loop_class == "none":
            return loop_code

        payload = {
            "code_snippet": loop_code,
            "cls": loop_class,
            "max_len": config.GENERATOR_MAX_LEN,
        }

        try:
            response = requests.post(
                self.generator_endpoint,
                json=payload,
                timeout=config.REQUEST_TIMEOUT_S,
            )
            if response.status_code != 200:
                raise RuntimeError(
                    f"Error while parallelizing, "
                    f"in generator with error: {response.text}"
                )
            pragma = response.json().get("pragma", "").strip()
            pragma = validate_pragma(pragma)
            return f"{pragma}\n{loop_code}"
        except Exception as e:
            raise RuntimeError(
                f"Error while parallelizing, in generator with error: {e}"
            )

    def parallelize(self, file_path: str) -> str:
        with open(file_path, "r") as file:
            code = file.read()

        if file_path.endswith(".py"):
            code = py2cpp_converter(code)

        loops = extract_loops(code)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        json_path = self.cache_dir / f"extracted_loops_{Path(file_path).stem}.json"

        if not os.path.exists(json_path):
            self.cache_dir.mkdir(parents=True, exist_ok=True)
            try:
                write_loops_to_json(loops, str(json_path))
            except Exception as e:
                raise RuntimeError("Error writing JSON") from e

        with open(json_path, "r") as f:
            loops_data = json.load(f)

        code_lines = code.split("\n")

        shift = 0
        for loop in loops_data:
            loop_code = loop["code"]
            start_line = loop["start_line"]
            end_line = loop["end_line"]
            loop_type = loop["type"]

            if loop_type != "for":
                continue

            features = extract_features(loop_code)
            embedding = vectorize_features(features)
            pred_class = self._classify(embedding)

            if pred_class != "none":
                pragma_with_loop = self._generate_omp_pragma_with_loop(
                    loop_code, pred_class
                )
                target_start = start_line - 1 + shift
                target_end = end_line + shift
                new_segment = pragma_with_loop.split("\n")
                code_lines[target_start:target_end] = new_segment
                current_segment_len = target_end - target_start
                new_segment_len = len(new_segment)
                shift += new_segment_len - current_segment_len

        current_dir = Path(file_path).parent
        final_output_path = current_dir / f"parallelized_{Path(file_path).stem}.c"

        with open(final_output_path, "w") as file:
            file.write("\n".join(code_lines))


parallelizer = Parallelizer()



import os

import matplotlib.pyplot as plt
import seaborn as sns

from accelera.src.automl.wrappers.tabular_graph_base import TabularGraphBase


class NumericalClassification(TabularGraphBase):
    def __init__(self, df, col_name, target_name, folder_path):
        super().__init__(df, col_name, target_name, folder_path)

    def build_graph(self):
        _, ax = plt.subplots(1, 3, figsize=(12, 4))
        # pie plot of nulls percent
        ax[0].pie(
            [float(self.nulls_percent), float(100 - self.nulls_percent)],
            labels=["Nulls", "Not Nulls"],
            autopct="%1.1f%%",
            colors=["#021D25", "#ADD8E6"],
        )
        ax[0].set_title(f"{self.col_name} Null percentage")
        sns.histplot(data=self.graph_df, x=self.col_name, ax=ax[1], kde=True)
        ax[1].set_title(f"{self.col_name} Distribution")
        ax[1].set_xlabel(self.col_name)
        ax[1].set_ylabel("Count")
        for label in ax[1].get_xticklabels():
            label.set_rotation(45)
            label.set_horizontalalignment("right")
        self.graph_df = self.graph_df[[self.col_name, self.target_name]].dropna()
        sns.boxplot(
            data=self.graph_df, x=self.target_name, y=self.col_name, ax=ax[2]
        )
        ax[2].set_title(f"{self.col_name} Distribution by {self.target_name}")
        ax[2].set_xlabel(self.target_name)
        ax[2].set_ylabel(self.col_name)
        for label in ax[2].get_xticklabels():
            label.set_rotation(45)
            label.set_horizontalalignment("right")
        plt.tight_layout()
        plt.savefig(os.path.join(self.folder_path, f"{self.col_name}.png"))
        plt.close()



import os

import matplotlib.pyplot as plt
import numpy as np

from accelera.src.automl.wrappers.graph_base import GraphBase


class SegmentationImagesAfterLoader(GraphBase):
    def __init__(
        self,
        images,
        masks,
        folder_path,
        title="",
        file_name="",
    ):
        super().__init__(folder_path)
        self.images = images
        self.masks = masks
        self.title = title
        self.file_name = file_name

    def build_graph(self):
        cols = len(self.images)
        fig, ax = plt.subplots(2, cols, figsize=(4 * cols, 4 * 2))
        if cols == 1:
            ax = np.array([[ax[0]], [ax[1]]])
        fig.suptitle(self.title, fontsize=20)
        for i, (img, mask) in enumerate(zip(self.images, self.masks)):
            img = img.permute(1, 2, 0).numpy()
            img = np.clip(img, 0, 1)
            ax[0][i].imshow(img)
            ax[0][i].set_title("image")
            ax[0][i].axis("off")
            mask = mask.numpy()
            if mask.ndim == 3:
                mask = mask[0]
            ax[1][i].imshow(mask, cmap="gray")

            ax[1][i].set_title("mask")
            ax[1][i].axis("off")

        plt.tight_layout()
        plt.savefig(os.path.join(self.folder_path, f"{self.file_name}.png"))
        plt.close()



import pandas as pd

from accelera.src.accelera_pipe.core.metric_display import MetricDisplay


class DisplayTupleNotCurve(MetricDisplay):
    def __init__(self, metric_name, values, folderpath):
        super().__init__(metric_name, values)
        self.folderpath = folderpath

    def execute(self):
        content = (
            "<div>\n"
            f"<h3>Metric name: {self.metric_name}</h3>\n"
            "<div class='metric-container'>\n"
        )

        for value in self.values:
            data = {}
            labels_name = self.handle_name("labels_name", len(value["result"][0]))
            headers_name = self.handle_name(
                "headers_name", len(self.values[0]["result"])
            )
            data["labels"] = labels_name
            for i in range(len(headers_name)):
                data[headers_name[i]] = value["result"][i]
            table = pd.DataFrame(data).to_html(
                index=False, border=1, justify="center"
            )
            new_content = (
                "<div>\n"
                "<h4>\n"
                f"Metric id :{value['metric id']}</h4>\n"
                f"{table}\n"
                "</div>\n"
            )
            content = content + new_content
        content = content + "</div>\n</div>\n"
        return content



import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline

from accelera.src.automl.wrappers.date_feature_extractor import DateFeatureExtractor
from accelera.src.automl.wrappers.flatten_1d_transform import Flatten1DTransform
from accelera.src.automl.wrappers.frequency_encoder_transform import (
    FrequencyEncoderTransform,
)
from accelera.src.automl.wrappers.IQR_transform import IQRTransform


class TestCustomTransform:
    def test_frequency_encoder_transform(self):
        data = pd.DataFrame(
            {
                "color": ["red", "blue", "red", "green", "blue", "blue"],
                "size": ["S", "L", "L", "S", "S", "L"],
            }
        )
        expected_output = pd.DataFrame(
            {
                "color": [
                    0.3333333333333333,
                    0.5,
                    0.3333333333333333,
                    0.16666666666666666,
                    0.5,
                    0.5,
                ],
                "size": [0.5, 0.5, 0.5, 0.5, 0.5, 0.5],
            }
        )
        transformer = FrequencyEncoderTransform()
        transformer.fit(data)
        transformed_data_data = transformer.transform(data)
        assert "color" in transformer.mapping_
        assert "size" in transformer.mapping_
        assert transformer.mapping_["color"] == {
            "blue": 0.5,
            "red": 0.3333333333333333,
            "green": 0.16666666666666666,
        }
        assert transformer.mapping_["size"] == {"S": 0.5, "L": 0.5}
        assert np.array_equal(transformed_data_data, expected_output.values)

    def test_frequency_encoder_transform_unseen_category(self):
        training_data = pd.DataFrame(
            {
                "color": ["red", "blue", "red", "green", "blue", "blue"],
                "size": ["S", "L", "L", "S", "S", "L"],
            }
        )
        testing_data = pd.DataFrame(
            {"color": ["red", "yellow"], "size": ["S", "XL"]}
        )
        expected_output = pd.DataFrame(
            {"color": [0.3333333333333333, 0], "size": [0.5, 0]}
        )
        transformer = FrequencyEncoderTransform()
        transformer.fit(training_data)
        transformed_data = transformer.transform(testing_data)
        assert np.array_equal(transformed_data, expected_output.values)

    def test_frequency_encoder_transform_pipeline(self):
        df = pd.DataFrame(
            {
                "color": ["red", "blue", "red", "green", "blue", "blue"],
                "size": ["S", "L", "L", "S", "S", "L"],
            }
        )
        expected_output = pd.DataFrame(
            {
                "color": [
                    0.3333333333333333,
                    0.5,
                    0.3333333333333333,
                    0.16666666666666666,
                    0.5,
                    0.5,
                ],
                "size": [0.5, 0.5, 0.5, 0.5, 0.5, 0.5],
            }
        )
        pipeline = Pipeline([("freq_encoder", FrequencyEncoderTransform())])
        transformer = ColumnTransformer(
            [("freq_encoder", pipeline, ["color", "size"])]
        )
        transformed_data = transformer.fit_transform(df)
        assert np.array_equal(transformed_data, expected_output.values)

    def test_flatten_1d_transform(self):
        data = np.array([["Hello"], ["World"]])
        expected_output = np.array(["Hello", "World"])
        transformer = Flatten1DTransform(func=lambda x: x.ravel())
        transformed_data = transformer.fit_transform(data)
        assert transformed_data.shape == expected_output.shape
        assert np.array_equal(transformed_data, expected_output)

    def test_flatten_1d_transform_pipeline(self):
        data = pd.DataFrame({"col": ["Hello", "World"]})

        expected_output = np.array([["Hello"], ["World"]])
        pipeline = Pipeline(
            [
                (
                    "flatten",
                    Flatten1DTransform(
                        func=lambda x: x.values.ravel()[:, np.newaxis]
                    ),
                )
            ]
        )
        transformer = ColumnTransformer([("flatten", pipeline, ["col"])])
        transformed_data = transformer.fit_transform(data)
        assert transformed_data.shape == expected_output.shape
        assert np.array_equal(transformed_data, expected_output)

    def test_iqr_transform(self):
        data = pd.DataFrame(
            {"col": [1, 2, 3, 4, 5, 100], "col2": [10, 20, 30, 40, 50, 1000]}
        )
        info = {
            "col": {"col_type": "numerical", "outliers_info": (2, 5)},
            "col2": {"col_type": "numerical", "outliers_info": (20, 50)},
        }
        expected_output = pd.DataFrame(
            {"col": [2, 2, 3, 4, 5, 5], "col2": [20, 20, 30, 40, 50, 50]}
        )
        transformer = IQRTransform(info=info, cols=["col", "col2"])
        output = transformer.fit_transform(data.values)
        assert np.array_equal(output, expected_output.values)

    def test_iqr_transform_pipeline_data_frame(self):
        data = pd.DataFrame(
            {"col": [1, 2, 3, 4, 5, 100], "col2": [10, 20, 30, 40, 50, 1000]}
        )
        info = {
            "col": {"col_type": "numerical", "outliers_info": (2, 5)},
            "col2": {"col_type": "numerical", "outliers_info": (20, 50)},
        }
        expected_output = pd.DataFrame(
            {"col": [2, 2, 3, 4, 5, 5], "col2": [20, 20, 30, 40, 50, 50]}
        )
        pipeline = Pipeline(
            [("iqr_transform", IQRTransform(info=info, cols=["col", "col2"]))]
        )
        transformer = ColumnTransformer(
            [("iqr_transform", pipeline, ["col", "col2"])]
        )
        output = transformer.fit_transform(data)
        assert np.array_equal(output, expected_output.values)

    def test_iqr_transform_pipeline_numpy(self):
        data = pd.DataFrame(
            {"col": [1, 2, 3, 4, 5, 100], "col2": [10, 20, 30, 40, 50, 1000]}
        )
        info = {
            "col": {"col_type": "numerical", "outliers_info": (2, 5)},
            "col2": {"col_type": "numerical", "outliers_info": (20, 50)},
        }
        expected_output = pd.DataFrame(
            {"col": [2, 2, 3, 4, 5, 5], "col2": [20, 20, 30, 40, 50, 50]}
        )
        pipeline = Pipeline(
            [("iqr_transform", IQRTransform(info=info, cols=["col", "col2"]))]
        )
        transformer = ColumnTransformer([("iqr_transform", pipeline, [0, 1])])
        output = transformer.fit_transform(data.values)
        assert np.array_equal(output, expected_output.values)

    def test_date_faeture_extractor(self):
        data = pd.DataFrame(
            {
                "date": pd.to_datetime(["2025-01-01", "2025-01-02", "2025-01-03"]),
                "num": [1, 2, 3],
            }
        )
        date_extractor = DateFeatureExtractor(cols=["date"])
        X = date_extractor.fit_transform(data)
        feature_names = date_extractor.get_feature_names_out()
        new_date = pd.DataFrame(X, columns=feature_names)
        assert feature_names == [
            "date_year",
            "date_month",
            "date_day",
            "date_weekday",
            "date_hour",
        ]
        corrected_data = pd.DataFrame(
            {
                "date_year": [2025, 2025, 2025],
                "date_month": [1, 1, 1],
                "date_day": [1, 2, 3],
                "date_weekday": [2, 3, 4],
                "date_hour": [0, 0, 0],
            }
        )
        pd.testing.assert_frame_equal(corrected_data, new_date)

    def test_date_faeture_extractor_pipeline_data_frame(self):
        data = pd.DataFrame(
            {
                "date": pd.to_datetime(["2025-01-01", "2025-01-02", "2025-01-03"]),
                "num": [1, 2, 3],
            }
        )
        corrected_data = pd.DataFrame(
            {
                "date_year": [2025, 2025, 2025],
                "date_month": [1, 1, 1],
                "date_day": [1, 2, 3],
                "date_weekday": [2, 3, 4],
                "date_hour": [0, 0, 0],
                "num": [1, 2, 3],
            }
        )
        pipeline = Pipeline([("date_extractor", DateFeatureExtractor(["date"]))])
        transformer = ColumnTransformer(
            [("date_extractor_col", pipeline, ["date"])],
            remainder="passthrough",
        )
        output = transformer.fit_transform(data)
        assert np.array_equal(output, corrected_data.values)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src.accelera_pipe import core as core
from accelera.src.accelera_pipe import wrappers as wrappers



import os
import textwrap

import matplotlib.pyplot as plt

from accelera.src.accelera_pipe.wrappers.graph_pipeline_report import (
    GraphPipelineReport,
)


class ModelReport(GraphPipelineReport):
    def __init__(self, folderpath, results, history=None):
        super().__init__(folderpath, results)
        if history and not isinstance(history, dict):
            raise TypeError(
                f"Expected 'history' to be a dict has "
                "the evaluation metrics of "
                f"the model to plot them, got {type(history).__name__}"
            )
        self.history = history
        self.metric_ids = list(map(str, range(len(results))))

    def display_history(self):
        try:
            content = "<h2> Model History</h2>\n"
            training_evaluation = []
            for key in self.history.keys():
                if not key.startswith("val_"):
                    training_evaluation.append(key)
            plt.figure(figsize=(8, 8))
            for i in range(len(training_evaluation)):
                plt.subplot(len(training_evaluation), 1, i + 1)
                plt.plot(
                    self.history[training_evaluation[i]],
                    label=training_evaluation[i],
                )
                plt.suptitle(f"{training_evaluation[i]}")
                val_name = f"val_{training_evaluation[i]}"
                validation = self.history[val_name]
                if validation:
                    plt.plot(validation, label=val_name)
                plt.xlabel("epochs")
                plt.legend()
            plt.tight_layout()
            path = os.path.join(self.folderpath, "history")
            plt.savefig(path)
            content += "<img src='history.png' alt='Model History' />\n"
            return content
        except Exception as e:
            print(f"Error while generating model history: {e}")
            return "<h2> Model History</h2>\nError: Could not generate plot.\n"

    def execute(self):
        metric_content = self.metric_display()
        content = textwrap.dedent(
            """\
        <p>This is the automated report for the 
        model  created using <i>Accelera</i>.</p>
        <p>
        It provides a comprehensive overview of the model’s 
        performance, including:</p>
        <ul>
        <li>
        Performance Summary — Highlighting key evaluation 
        metrics and model results</li>
        <li>
        - Training Graphs — Visualizing deep learning history 
        (loss, accuracy, and other tracked metrics) if exists
        </li>
        </ul>
        """
        )
        if self.history:
            content += self.display_history()
        content = (
            self.start_content
            + "\n"
            + content
            + "\n"
            + metric_content
            + self.end_content
        )
        self.create_html_file(content)



from unittest.mock import patch

import numpy as np
import pytest
import torch
from PIL import Image

from accelera.src.automl.core.segmentation_image_dataset import (
    SegmentationImageDataset,
)


class TestSegmentationImageDataset:
    @pytest.fixture(autouse=True)
    def create_sample_data(self, tmp_path):
        self.paths = []
        self.mask_paths = []
        for i in range(5):
            path = tmp_path / f"img{i}.png"
            img = np.random.randint(0, 256, (256, 256, 3), dtype=np.uint8)
            img = Image.fromarray(img)
            img.save(path)
            self.paths.append(str(path))
            mask_path = tmp_path / f"mask{i}.png"
            mask = np.random.randint(0, 256, (256, 256), dtype=np.uint8)
            mask = Image.fromarray(mask)
            mask.save(mask_path)
            self.mask_paths.append(str(mask_path))

    def test___getitem__(self):
        dataset = SegmentationImageDataset(
            image_paths=self.paths,
            masks=self.mask_paths,
            augment=False,
            image_size=(125, 125),
        )
        img_tensor, mask_tensor = dataset[0]
        assert len(dataset) == len(self.paths)
        assert img_tensor.shape[0] == 3
        assert img_tensor.shape[1] == 125
        assert img_tensor.shape[2] == 125
        assert mask_tensor.shape[0] == 1
        assert mask_tensor.shape[1] == 125
        assert mask_tensor.shape[2] == 125
        assert torch.all((mask_tensor == 1) | (mask_tensor == 0))
        assert isinstance(img_tensor, torch.Tensor)
        assert isinstance(mask_tensor, torch.Tensor)
        dataset = SegmentationImageDataset(
            image_paths=self.paths,
            masks=None,
            augment=False,
            image_size=(125, 125),
        )
        img_tensor, mask_tensor = dataset[0]
        assert len(dataset) == len(self.paths)
        assert img_tensor.shape[0] == 3
        assert img_tensor.shape[1] == 125
        assert img_tensor.shape[2] == 125
        assert mask_tensor is None
        assert isinstance(img_tensor, torch.Tensor)

    def test_random_horizontal_flip(self):
        img = Image.open(self.paths[0]).convert("RGB")
        mask = Image.open(self.mask_paths[0]).convert("L")
        dataset = SegmentationImageDataset(
            [], [], horizontal_flip=False, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_horizontal_flip(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], horizontal_flip=True, augmentation_probability=0
        )
        returned_img, returned_mask = dataset.random_horizontal_flip(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], horizontal_flip=True, augmentation_probability=1
        )
        with patch("random.random", return_value=0.0):
            returned_img, returned_mask = dataset.random_horizontal_flip(img, mask)
            returned_img2, returned_mask2 = dataset.random_horizontal_flip(
                returned_img, returned_mask
            )
        assert np.array_equal(np.array(img), np.array(returned_img2))
        assert np.array_equal(np.array(mask), np.array(returned_mask2))
        dataset = SegmentationImageDataset(
            [], [], horizontal_flip=True, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_horizontal_flip(img, None)
        assert not np.array_equal(np.array(img), np.array(returned_img))
        assert returned_mask is None

    def test_random_vertical_flip(self):
        img = Image.open(self.paths[0]).convert("RGB")
        mask = Image.open(self.mask_paths[0]).convert("L")

        dataset = SegmentationImageDataset(
            [], [], vertical_flip=False, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_vertical_flip(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))

        dataset = SegmentationImageDataset(
            [], [], vertical_flip=True, augmentation_probability=0
        )
        returned_img, returned_mask = dataset.random_vertical_flip(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], vertical_flip=True, augmentation_probability=1
        )
        with patch("random.random", return_value=0.0):
            returned_img, returned_mask = dataset.random_vertical_flip(img, mask)
            returned_img2, returned_mask2 = dataset.random_vertical_flip(
                returned_img, returned_mask
            )
        assert np.array_equal(np.array(img), np.array(returned_img2))
        assert np.array_equal(np.array(mask), np.array(returned_mask2))
        dataset = SegmentationImageDataset(
            [], [], vertical_flip=True, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_vertical_flip(img, None)
        assert not np.array_equal(np.array(img), np.array(returned_img))
        assert returned_mask is None

    def test_random_rotation(self):
        img = Image.open(self.paths[0]).convert("RGB")
        mask = Image.open(self.mask_paths[0]).convert("L")

        dataset = SegmentationImageDataset(
            [], [], rotation=False, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_rotation(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], rotation=True, augmentation_probability=0
        )
        returned_img, returned_mask = dataset.random_rotation(img, mask)
        assert np.array_equal(np.array(img), np.array(returned_img))
        assert np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], rotation=True, augmentation_probability=1
        )
        with (
            patch("random.random", return_value=0.0),
            patch("random.uniform", return_value=30.0),
        ):
            returned_img, returned_mask = dataset.random_rotation(img, mask)
        assert not np.array_equal(np.array(img), np.array(returned_img))
        assert not np.array_equal(np.array(mask), np.array(returned_mask))
        dataset = SegmentationImageDataset(
            [], [], rotation=True, augmentation_probability=1
        )
        returned_img, returned_mask = dataset.random_rotation(img, None)
        assert not np.array_equal(np.array(img), np.array(returned_img))
        assert returned_mask is None



import numpy as np
import pytest
from PIL import Image

from accelera.src.automl.core.classification_image_testing_preprocessing import (  # noqa: E501
    ClassificationImageTestingPreprocessing,
)
from accelera.src.utils.preprocessing import save_pickle


class TestClassificationImageTestingPreprocessing:
    def test_constructor(self, tmp_path):
        data_info = {"image_size": (125, 125)}
        save_pickle(tmp_path, data_info, "data_info.pkl")
        class2label_mapping = {"cats": 0, "dogs": 1}
        save_pickle(tmp_path, class2label_mapping, "class2label_mapping.pkl")
        with pytest.raises(
            ValueError, match="Image paths must be list of paths not none"
        ):
            ClassificationImageTestingPreprocessing(
                image_paths=None, folder_path=tmp_path
            )

        with pytest.raises(ValueError, match="Image paths must be list of paths"):
            ClassificationImageTestingPreprocessing(
                image_paths="path", folder_path=tmp_path
            )
        with pytest.raises(ValueError, match="Image paths is empty list"):
            ClassificationImageTestingPreprocessing(
                image_paths=[], folder_path=tmp_path
            )

        with pytest.raises(
            ValueError, match="Class names must be list of class names"
        ):
            ClassificationImageTestingPreprocessing(
                image_paths=["path"], image_class_names=0, folder_path=tmp_path
            )

        with pytest.raises(
            ValueError, match="Image paths length must equal class names length"
        ):
            ClassificationImageTestingPreprocessing(
                image_paths=["path1", "path2"],
                image_class_names=[0],
                folder_path=tmp_path,
            )
        with pytest.raises(ValueError, match="There is no valid image exists"):
            ClassificationImageTestingPreprocessing(
                image_paths=["path1", "path2"],
                image_class_names=[0, 1],
                folder_path=tmp_path,
            )
        paths = []
        for i in range(5):
            img = Image.fromarray(
                np.random.randint(0, 256, (225, 225, 3), dtype=np.uint8)
            )
            img.save(tmp_path / f"img_{i}.png")
            paths.append(str(tmp_path / f"img_{i}.png"))
        preprocesseor = ClassificationImageTestingPreprocessing(
            image_paths=paths,
            image_class_names=None,
            folder_path=tmp_path,
        )
        assert len(preprocesseor.valid_images) == 5

    def test_common_preprocessing(self, tmp_path):
        data_info = {"image_size": (125, 125)}
        save_pickle(tmp_path, data_info, "data_info.pkl")
        class2label_mapping = {"cats": 0, "dogs": 1}
        save_pickle(tmp_path, class2label_mapping, "class2label_mapping.pkl")
        paths = []
        for i in range(5):
            img = Image.fromarray(
                np.random.randint(0, 256, (225, 225, 3), dtype=np.uint8)
            )
            img.save(tmp_path / f"img_{i}.png")
            paths.append(str(tmp_path / f"img_{i}.png"))
        testing_loader, invalid_paths = ClassificationImageTestingPreprocessing(
            image_paths=paths,
            image_class_names=None,
            folder_path=tmp_path,
        ).common_preprocessing()
        assert testing_loader is not None
        assert len(invalid_paths) == 0
        images = next(iter(testing_loader))
        assert len(images) == 4
        labels = ["cats", "dogs", "cats", "birds", "dogs"]
        with pytest.raises(
            ValueError, match="this class name not in the training class"
        ):
            testing_loader, invalid_paths = ClassificationImageTestingPreprocessing(
                image_paths=paths,
                image_class_names=labels,
                folder_path=tmp_path,
            ).common_preprocessing()
        labels = ["cats", "dogs", "cats", "dogs", "dogs"]
        testing_loader, invalid_paths = ClassificationImageTestingPreprocessing(
            image_paths=paths,
            image_class_names=labels,
            folder_path=tmp_path,
        ).common_preprocessing()
        assert testing_loader is not None
        assert len(invalid_paths) == 0
        batch = next(iter(testing_loader))
        assert len(batch) == 2
        assert len(batch[0]) == 4
        assert len(batch[1]) == 4



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src.automl.wrappers import IQR_transform as IQR_transform
from accelera.src.automl.wrappers import (
    categorical_classification as categorical_classification,
)
from accelera.src.automl.wrappers import (
    categorical_regression as categorical_regression,
)
from accelera.src.automl.wrappers import (
    classification_images_after_loader as classification_images_after_loader,
)
from accelera.src.automl.wrappers import correlation_graph as correlation_graph
from accelera.src.automl.wrappers import (
    custom_transform_test as custom_transform_test,
)
from accelera.src.automl.wrappers import (
    date_feature_extractor as date_feature_extractor,
)
from accelera.src.automl.wrappers import (
    display_sample_images_classification as display_sample_images_classification,
)
from accelera.src.automl.wrappers import (
    display_sample_images_segmentation as display_sample_images_segmentation,
)
from accelera.src.automl.wrappers import flatten_1d_transform as flatten_1d_transform
from accelera.src.automl.wrappers import (
    frequency_encoder_transform as frequency_encoder_transform,
)
from accelera.src.automl.wrappers import graph_base as graph_base
from accelera.src.automl.wrappers import (
    image_label_classification as image_label_classification,
)
from accelera.src.automl.wrappers import (
    image_preprocessing_report as image_preprocessing_report,
)
from accelera.src.automl.wrappers import (
    numerical_classification as numerical_classification,
)
from accelera.src.automl.wrappers import numerical_regression as numerical_regression
from accelera.src.automl.wrappers import (
    ordinal_classification as ordinal_classification,
)
from accelera.src.automl.wrappers import ordinal_regression as ordinal_regression
from accelera.src.automl.wrappers import (
    segmentation_data_summary as segmentation_data_summary,
)
from accelera.src.automl.wrappers import (
    segmentation_images_after_loader as segmentation_images_after_loader,
)
from accelera.src.automl.wrappers import tabular_graph_base as tabular_graph_base
from accelera.src.automl.wrappers import (
    tabular_preprocessing_report as tabular_preprocessing_report,
)
from accelera.src.automl.wrappers import (
    target_classification as target_classification,
)
from accelera.src.automl.wrappers import target_regression as target_regression
from accelera.src.automl.wrappers import text_graph as text_graph



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src.automl import core as core
from accelera.src.automl import wrappers as wrappers



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src.accelera_pipe.wrappers import (
    display_array_multi as display_array_multi,
)
from accelera.src.accelera_pipe.wrappers import (
    display_array_single as display_array_single,
)
from accelera.src.accelera_pipe.wrappers import display_dict as display_dict
from accelera.src.accelera_pipe.wrappers import display_figure as display_figure
from accelera.src.accelera_pipe.wrappers import (
    display_single_number as display_single_number,
)
from accelera.src.accelera_pipe.wrappers import display_string as display_string
from accelera.src.accelera_pipe.wrappers import (
    display_tuple_not_curve as display_tuple_not_curve,
)
from accelera.src.accelera_pipe.wrappers import (
    graph_pipeline_report as graph_pipeline_report,
)
from accelera.src.accelera_pipe.wrappers import graph_report as graph_report
from accelera.src.accelera_pipe.wrappers import model_report as model_report
from accelera.src.accelera_pipe.wrappers import (
    supervised_metric as supervised_metric,
)
from accelera.src.accelera_pipe.wrappers import (
    unsupervised_metric as unsupervised_metric,
)



import os

import matplotlib.pyplot as plt

from accelera.src.accelera_pipe.core.metric_display import MetricDisplay
from accelera.src.utils.accelera_utils import create_folder

plt.style.use("dark_background")


class DisplayFigure(MetricDisplay):
    def __init__(self, metric_name, values, folderpath):
        super().__init__(metric_name, values)
        self.folderpath = folderpath

    def execute(self):
        content = (
            "<div>\n"
            f"<h3>Metric name: {self.metric_name}</h3>\n"
            "<div class='metric-container'>\n"
        )
        sub_folder_path = os.path.join(self.folderpath, self.metric_name)
        create_folder(sub_folder_path)
        for value in self.values:
            plot_func = value["plot_func"]
            result = value["result"]
            img_path = os.path.join(sub_folder_path, f"{value['metric id']}.png")
            plt = plot_func(result)
            if plt is None:
                raise ValueError("The plot_func must return the plt object")
            plt.title(f"{self.metric_name} - Metric ID: {value['metric id']}")
            plt.savefig(img_path)
            plt.close()
            img_src = os.path.join(self.metric_name, f"{value['metric id']}.png")
            img_alt = f"{self.metric_name}_{value['metric id']}"
            new_content = f"<div>\n<img src='{img_src}' alt='{img_alt}' />\n</div>\n"
            content = content + new_content
        content = content + "</div>\n</div>\n"
        return content



from accelera.src.custom.transformer import CustomTransformer


class Flatten1DTransform(CustomTransformer):
    def __init__(self, func):
        self.func = func

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return self.func(X)

    def get_feature_names_out(self, input_features=None):
        return input_features



import shutil
import tempfile

import pandas as pd
import pytest

from accelera.src.automl.core.image_training_preprocessing import (
    ImageTrainingPreprocessing,
)


class TestImageTrainingPreprocessing:
    @pytest.fixture(autouse=True)
    def temp_folder(self):
        self.temp_dir = tempfile.mkdtemp()
        yield self.temp_dir
        shutil.rmtree(self.temp_dir)

    def test_image_training_preprocessing_initialization(self, tmp_path):
        training_folder = tmp_path / "training"
        training_folder.mkdir()
        validation_folder = tmp_path / "validation"
        validation_folder.mkdir()
        with pytest.raises(ValueError, match="Training folder must be not null"):
            ImageTrainingPreprocessing(
                training_folder=None,
                validation_folder=None,
                split_training=False,
                random_state=42,
                images_size=(224, 224),
                val_size=0.2,
                folder_path=self.temp_dir,
            )

        with pytest.raises(ValueError):
            ImageTrainingPreprocessing(
                training_folder="invalid",
                validation_folder=None,
                split_training=False,
                random_state=42,
                images_size=(224, 224),
                val_size=0.2,
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder="invalid",
                split_training=False,
                random_state=42,
                images_size=(224, 224),
                val_size=0.2,
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=True,
                random_state=42,
                images_size=(224, 224),
                val_size=0.2,
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=training_folder,
                split_training=False,
                random_state=42,
                images_size=(224, 224),
                val_size=0.2,
                folder_path=self.temp_dir,
            )
        with pytest.raises(
            ValueError, match="Test size is invalid it must be less than 0.5"
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size="invalid type",
                random_state=42,
                images_size=(224, 224),
                folder_path=self.temp_dir,
            )
        with pytest.raises(
            ValueError, match="Test size is invalid it must be less than 0.5"
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.9,
                random_state=42,
                images_size=(224, 224),
                folder_path=self.temp_dir,
            )
        with pytest.raises(
            ValueError,
            match="Random state is invalid it must be integer or None",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42.4,
                images_size=(224, 224),
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError, match="Image size must be tuple"):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size="invalid",
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError, match="Image size is not integer"):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(224, 224.3),
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError, match="Image size is not integer"):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(224.5, 224),
                folder_path=self.temp_dir,
            )
        with pytest.raises(
            ValueError,
            match=r".*than or equal 32 and less than or equal 1024",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(16, 224),
                folder_path=self.temp_dir,
            )
        with pytest.raises(
            ValueError,
            match="batch_size must be a integer",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size="invalid",
            )
        with pytest.raises(
            ValueError,
            match="batch_size must be a positive value",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=-1,
            )
        with pytest.raises(
            ValueError,
            match="augment must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment="invalid",
            )
        with pytest.raises(
            ValueError,
            match=r"horizontal_flip must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                horizontal_flip="invalid",
            )
        with pytest.raises(
            ValueError,
            match="vertical_flip must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                vertical_flip="invalid",
            )
        with pytest.raises(
            ValueError,
            match="rotation must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                rotation="invalid",
            )
        with pytest.raises(
            ValueError,
            match="brightness must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                brightness="invalid",
            )
        with pytest.raises(
            ValueError,
            match="contrast must be a boolean",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                contrast="invalid",
            )
        with pytest.raises(
            ValueError,
            match=r"augmentation_probability must be between \[0,1\]",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                augmentation_probability="invalid",
            )
        with pytest.raises(
            ValueError,
            match=r"augmentation_probability must be between \[0,1\]",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                augmentation_probability=2,
            )
        with pytest.raises(
            ValueError,
            match="rotation_angle must be positive integer or float",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                rotation_angle="invalid",
            )
        with pytest.raises(
            ValueError,
            match="rotation_angle must be positive integer or float",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                rotation_angle=-1,
            )
        with pytest.raises(
            ValueError,
            match="brightness_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                brightness_factors=1,
            )
        with pytest.raises(
            ValueError,
            match="brightness_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                brightness_factors=(2),
            )
        with pytest.raises(
            ValueError,
            match="brightness_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                brightness_factors=(2, 3, 4),
            )
        with pytest.raises(
            ValueError,
            match="brightness_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                brightness_factors=("invalid", 4),
            )

        with pytest.raises(
            ValueError,
            match="contrast_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                contrast_factors=1,
            )
        with pytest.raises(
            ValueError,
            match="contrast_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                contrast_factors=(2),
            )
        with pytest.raises(
            ValueError,
            match="contrast_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                contrast_factors=(2, 3, 4),
            )
        with pytest.raises(
            ValueError,
            match="contrast_factors must be tuple of two items float or integers",
        ):
            ImageTrainingPreprocessing(
                training_folder=training_folder,
                validation_folder=validation_folder,
                split_training=False,
                val_size=0.2,
                random_state=42,
                images_size=(32, 224),
                folder_path=self.temp_dir,
                batch_size=16,
                augment=True,
                contrast_factors=("invalid", 4),
            )

    def test_get_sample_random(self, tmp_path):
        training_folder = tmp_path / "training"
        training_folder.mkdir()
        training_preprocessor = ImageTrainingPreprocessing(
            training_folder=training_folder,
            validation_folder=None,
            folder_path=self.temp_dir,
            split_training=False,
            images_size=(224, 224),
            random_state=42,
            val_size=0.2,
        )
        paths = ["image1.png", "image2.png"]
        labels = [0, 1]
        df = training_preprocessor.get_sample_random(
            "training", paths, labels, num_samples=2
        )
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 2
        assert list(df.columns) == ["training_paths", "training_labels"]

    def test_splitting(self, tmp_path):
        training_folder = tmp_path / "training"
        training_folder.mkdir()
        training_preprocessor = ImageTrainingPreprocessing(
            training_folder=training_folder,
            validation_folder=None,
            folder_path=self.temp_dir,
            split_training=False,
            images_size=(224, 224),
            random_state=42,
            val_size=0.2,
        )
        training_folder_images_paths = [
            "imag1",
            "imag2",
            "imag3",
            "imag4",
            "imag5",
            "imag6",
            "imag7",
            "imag8",
            "imag9",
            "imag10",
        ]
        validation_folder_images_paths = ["imag11", "imag12"]
        training_folder_images_labels = [0, 1, 1, 1, 0, 0, 0, 1, 1, 1]
        validation_folder_images_labels = [1, 0]
        training_preprocessor.splitting(
            training_folder_images_paths,
            validation_folder_images_paths,
            training_folder_images_labels,
            validation_folder_images_labels,
        )
        assert len(training_preprocessor.training_paths) == 10
        assert training_preprocessor.training_paths == training_folder_images_paths
        assert len(training_preprocessor.training_labels) == 10
        assert training_preprocessor.training_labels == training_folder_images_labels
        assert len(training_preprocessor.validation_paths) == 2
        assert (
            training_preprocessor.validation_paths == validation_folder_images_paths
        )
        assert len(training_preprocessor.validation_labels) == 2
        assert (
            training_preprocessor.validation_labels
            == validation_folder_images_labels
        )
        training_preprocessor = ImageTrainingPreprocessing(
            training_folder=training_folder,
            validation_folder=None,
            folder_path=self.temp_dir,
            split_training=True,
            images_size=(224, 224),
            random_state=42,
            val_size=0.2,
        )
        training_preprocessor.splitting(
            training_folder_images_paths,
            validation_folder_images_paths,
            training_folder_images_labels,
            validation_folder_images_labels,
        )
        assert len(training_preprocessor.training_paths) == 8
        assert len(training_preprocessor.training_labels) == 8
        assert len(training_preprocessor.validation_paths) == 2
        assert len(training_preprocessor.validation_labels) == 2



import os

import matplotlib.pyplot as plt


class GraphBase:
    def __init__(self, folder_path):
        self.folder_path = os.path.join(folder_path, "graphs")
        if not os.path.exists(self.folder_path):
            os.makedirs(self.folder_path)
        plt.style.use("dark_background")

    def build_graph(self):
        raise NotImplementedError("Subclasses should implement this method")



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from accelera.src import config as config
from accelera.src.version import __version__ as __version__
from accelera.src import accelera_pipe as accelera_pipe
from accelera.src import automl as automl
from accelera.src import benchmark as benchmark
from accelera.src import custom as custom
from accelera.src import utils as utils



import random

import numpy as np
import torch
from PIL import Image

from accelera.src.automl.core.image_dataset import ImageDataset


class ClassificationImageDataset(ImageDataset):
    def __init__(
        self,
        image_paths,
        labels=None,
        image_size=(224, 224),
        augment=False,
        augmentation_probability=0.5,
        horizontal_flip=False,
        vertical_flip=False,
        rotation=False,
        rotation_angle=30,
        brightness=False,
        brightness_factors=(0.8, 1.2),
        contrast=False,
        contrast_factors=(0.8, 1.2),
    ):
        super().__init__(
            image_paths,
            labels,
            image_size,
            augment,
            augmentation_probability,
            horizontal_flip,
            vertical_flip,
            rotation,
            rotation_angle,
            brightness,
            brightness_factors,
            contrast,
            contrast_factors,
        )

    def load_image(self, index):
        path = self.image_paths[index]
        img = Image.open(path).convert("RGB")
        img = img.resize(self.image_size)
        if self.augment:
            img = self.augmentation(img)
        img_array = np.array(img, dtype=np.float32) / 255.0
        img_transpose = np.transpose(img_array, (2, 0, 1))
        img_tensor = torch.tensor(img_transpose, dtype=torch.float32)
        return img_tensor

    def random_horizontal_flip(self, img):
        if self.horizontal_flip and random.random() < self.augmentation_probability:
            return img.transpose(Image.FLIP_LEFT_RIGHT)
        return img

    def random_vertical_flip(self, img):
        if self.vertical_flip and random.random() < self.augmentation_probability:
            return img.transpose(Image.FLIP_TOP_BOTTOM)
        return img

    def random_rotation(self, img):
        if self.rotation and random.random() < self.augmentation_probability:
            random_angle = random.uniform(
                -1 * self.rotation_angle, self.rotation_angle
            )
            return img.rotate(random_angle)
        return img

    def augmentation(self, img):
        img = self.random_horizontal_flip(img)
        img = self.random_vertical_flip(img)
        img = self.random_rotation(img)
        img = self.random_brightness(img)
        img = self.random_contrast(img)
        return img

    def __getitem__(self, index):
        img_tensor = self.load_image(index)
        label_tensor = None
        if self.labels is not None:
            label_tensor = torch.tensor(self.labels[index], dtype=torch.long)
        return img_tensor, label_tensor



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from accelera.src.automl.wrappers.graph_base import GraphBase


class DisplaySampleImagesClassification(GraphBase):
    def __init__(
        self,
        paths,
        labels,
        label2class_mapping,
        folder_path,
        n_sample=4,
        title="",
        file_name="",
    ):
        super().__init__(folder_path)
        self.label2class_mapping = label2class_mapping
        self.title = title
        self.file_name = file_name
        self.n_sample = n_sample
        classes = list(self.label2class_mapping.keys())
        self.df = pd.DataFrame({"paths": paths, "labels": labels})
        self.temp_df = {}
        for class_idx in classes:
            class_paths = self.df[self.df["labels"] == class_idx]
            class_df = class_paths.sample(
                n=min(self.n_sample, len(class_paths)), random_state=42
            )
            self.temp_df[label2class_mapping[class_idx]] = class_df["paths"]

    def build_graph(self):
        classes = list(self.temp_df.keys())
        num_rows = len(classes)
        fig, ax = plt.subplots(
            num_rows, self.n_sample, figsize=(4 * self.n_sample, 4 * num_rows)
        )
        fig.suptitle(self.title, fontsize=20)
        ax = np.atleast_2d(ax)
        for row, class_name in enumerate(classes):
            paths = self.temp_df[class_name]
            for col, path in enumerate(paths):
                ax[row][col].axis("off")
                img = plt.imread(path)
                ax[row][col].imshow(img)
                ax[row][col].set_title(class_name)
        plt.tight_layout()
        plt.savefig(os.path.join(self.folder_path, f"{self.file_name}.png"))
        plt.close()



from __future__ import annotations

import os
import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path


def _repo_root_from_this_file() -> Path:
    return Path(__file__).resolve().parents[2]


@dataclass(frozen=True)
class Config:
    REPO_ROOT: Path = _repo_root_from_this_file()
    CACHE_DIR_NAME: str = ".accelera_cache"
    BUILD_DIR_NAME: str = "build"
    BINDINGS_SUBDIR: str = "bindings"
    API_SUBDIR: str = "api"
    DATASETS: dict[str, dict] = field(
        default_factory=lambda: {
            "customer_purchase_data": {
                "id": "1OCe4X8mVFEwWz0cfQyAJKNiZMbiUGbB2",
                "metadata": {},
            },
            "TestReviews": {
                "id": "1pBNGL9NuScgUb16Yh8W6gFzhXanCWa9q",
                "metadata": {},
            },
            "Dataset_spine": {
                "id": "1RmaNadcroXrqodMq1APBQBhIKBIZLbvJ",
                "metadata": {},
            },
            "DailyDialog": {
                "id": "1oWgWAZpvVAgnxK3P-0sqElWDWMozdK4P",
                "metadata": {},
            },
            "job_salary_prediction_dataset": {
                "id": "1r5HEDyelbzfZbXPSx-0G7vEvrgXGqb9r",
                "metadata": {},
            },
            "Housing": {
                "id": "1VMtLcWDcigwkimpf-eWVMZ7zJMUf7wxs",
                "metadata": {},
            },
            "student_placement_synthetic": {
                "id": "1l3YRaiPNA4EowBFuj6cm0v4yS2DQOY1M",
                "metadata": {},
            },
            "Titanic-Dataset": {
                "id": "1VWRzvPDcFOo7YByJtKKupH3H0czFAHCx",
                "metadata": {},
            },
            "heart": {
                "id": "1HG9BPIS-PFHzzXBnKrohMCrmr7W9oxJY",
                "metadata": {},
            },
            "student_exam_performance_dataset": {
                "id": "19mH4wgsK2-zcttdTFl6jk_BR83lrKIS7",
                "metadata": {},
            },
        }
    )

    DEFAULT_CLANG_ARGS: tuple[str, ...] = ("-std=c++17",)
    CLANG_FORMAT_BIN: str = os.getenv("ACCELERA_CLANG_FORMAT_BIN", "clang-format")
    ENABLE_CPP_FORMATTING: bool = (
        os.getenv("ACCELERA_ENABLE_CPP_FORMATTING", "1") == "1"
    )

    CLASSIFIER_ENDPOINT: str = os.getenv(
        "ACCELERA_CLASSIFIER_ENDPOINT",
        "https://accelera-ai-open-mp-classifier.hf.space/predict",
    )
    GENERATOR_ENDPOINT: str = os.getenv(
        "ACCELERA_GENERATOR_ENDPOINT",
        "https://accelera-ai-open-mp-generator.hf.space/generate",
    )
    DATASET_DRIVE_FOLDER_ENDPOINT: str = (
        "https://drive.google.com/drive/folders/1ijBkIdvLnVGbINjDklTsC3-z3jc-_GIE"
    )
    GENERATOR_MAX_LEN: int = int(os.getenv("ACCELERA_GENERATOR_MAX_LEN", "1500"))
    REQUEST_TIMEOUT_S: int = int(os.getenv("ACCELERA_REQUEST_TIMEOUT_S", "10"))

    # If 1, `accelera/__init__.py` will
    # raise if optional `accelera.api` import fails.
    STRICT_API_IMPORT: bool = os.getenv("ACCELERA_STRICT_API_IMPORT", "0") == "1"

    @property
    def cache_dir(self) -> Path:
        return self.REPO_ROOT / self.CACHE_DIR_NAME

    @property
    def bindings_dir(self) -> Path:
        return self.REPO_ROOT / self.BUILD_DIR_NAME / self.BINDINGS_SUBDIR

    @property
    def api_dir(self) -> Path:
        return self.REPO_ROOT / "accelera" / self.API_SUBDIR

    def ensure_bindings_on_syspath(self) -> None:
        bindings = str(self.bindings_dir)
        if bindings not in sys.path:
            sys.path.insert(0, bindings)


config = Config()



import os

from torch.utils.data import DataLoader

from accelera.src.automl.core.classification_image_dataset import (
    ClassificationImageDataset,
)
from accelera.src.automl.core.image_training_preprocessing import (
    ImageTrainingPreprocessing,
)
from accelera.src.automl.wrappers.classification_images_after_loader import (
    ClassificationImagesAfterLoader,
)
from accelera.src.automl.wrappers.display_sample_images_classification import (
    DisplaySampleImagesClassification,
)
from accelera.src.automl.wrappers.image_label_classification import (
    ImageLabelClassification,
)
from accelera.src.automl.wrappers.image_preprocessing_report import (
    ImagePreprocessingReport,
)
from accelera.src.utils.preprocessing import get_sub_folders_names
from accelera.src.utils.preprocessing import is_valid_image
from accelera.src.utils.preprocessing import save_pickle


class ClassificationImageTrainingPreprocessing(ImageTrainingPreprocessing):
    def __init__(
        self,
        training_folder_images,
        folder_path,
        validation_folder_images=None,
        split_training=False,
        val_size=0.2,
        random_state=42,
        images_size=(224, 224),
        augment=True,
        batch_size=16,
        augmentation_probability=0.5,
        horizontal_flip=True,
        vertical_flip=True,
        rotation=True,
        rotation_angle=30,
        brightness=True,
        brightness_factors=(0.8, 1.2),
        contrast=True,
        contrast_factors=(0.8, 1.2),
    ):
        super().__init__(
            training_folder_images,
            folder_path,
            validation_folder_images,
            split_training,
            val_size,
            random_state,
            images_size,
            augment,
            batch_size,
            augmentation_probability,
            horizontal_flip,
            vertical_flip,
            rotation,
            rotation_angle,
            brightness,
            brightness_factors,
            contrast,
            contrast_factors,
        )
        self.training_class = get_sub_folders_names(self.training_folder)
        self.validation_class = None
        self.training_folder_invalid_images = []
        self.training_folder_invalid_images_labels = []
        self.validation_folder_invalid_images = []
        self.validation_folder_invalid_images_labels = []
        if len(self.training_class) == 0:
            raise ValueError("Training Folder dosen't have any folder inside it ")
        if self.validation_folder is not None:
            self.validation_class = get_sub_folders_names(self.validation_folder)
            if len(self.validation_class) == 0:
                raise ValueError(
                    "Validation Folder dosen't have any folder inside it "
                )
            for class_name in self.validation_class:
                if class_name not in self.training_class:
                    raise ValueError(
                        f"This category {class_name} not in the training "
                        f"categories which are {self.training_class}"
                    )
        (
            self.validation_folder_images_paths,
            self.validation_folder_images_labels,
        ) = (
            None,
            None,
        )
        data_info = {"image_size": self.image_size}
        save_pickle(self.folder_path, data_info, "data_info.pkl")

    def get_classes_mapping(self):
        self.class2label_mapping = {}
        self.label2class_mapping = {}
        for idx, class_name in enumerate(sorted(self.training_class)):
            self.class2label_mapping[class_name] = idx
            self.label2class_mapping[idx] = class_name
        save_pickle(
            self.folder_path,
            self.class2label_mapping,
            "class2label_mapping.pkl",
        )
        save_pickle(
            self.folder_path,
            self.label2class_mapping,
            "label2class_mapping.pkl",
        )

    def data_preparing(
        self, folder_path, invalid_list_paths, invalid_list_labels, classes_name
    ):
        paths = []
        labels = []
        for class_name in classes_name:
            sub_folder_path = os.path.join(folder_path, class_name)
            for path in os.listdir(sub_folder_path):
                path = os.path.join(sub_folder_path, path)
                mapping = self.class2label_mapping[class_name]
                if is_valid_image(path):
                    paths.append(path)
                    labels.append(mapping)
                else:
                    invalid_list_paths.append(path)
                    invalid_list_labels.append(mapping)
        if len(paths) == 0:
            raise ValueError("There is no valid path")
        return paths, labels

    def data_overview(self):
        train_df = self.get_sample_random(
            "training_folder",
            self.training_folder_images_paths,
            self.training_folder_images_labels,
        )
        self.report_data["data_overview"] = {}
        self.report_data["data_overview"]["training_folder"] = {
            "classes": self.training_class,
            "images_len": len(self.training_folder_images_paths),
            "random_sample": train_df.head(),
            "invalid_len": len(self.training_folder_invalid_images),
            "invalid_images": self.training_folder_invalid_images,
            "mapping": self.class2label_mapping,
        }

        if self.validation_folder is not None:
            val_df = self.get_sample_random(
                "validation_folder",
                self.validation_folder_images_paths,
                self.validation_folder_images_labels,
            )
            self.report_data["data_overview"]["validation_folder"] = {
                "classes": self.validation_class,
                "images_len": len(self.validation_folder_images_paths),
                "random_sample": val_df.head(),
                "invalid_len": len(self.validation_folder_invalid_images),
                "invalid_images": self.validation_folder_invalid_images,
            }

    def make_graphs_label_summary(self):
        ImageLabelClassification(
            self.training_folder_images_labels,
            self.label2class_mapping,
            self.training_folder_invalid_images_labels,
            self.folder_path,
            "Training Folder Labels Summary ",
            "training_folder_labels_summary",
        ).build_graph()
        self.report_data["graphs"]["images_name"].append(
            "training_folder_labels_summary"
        )
        if self.validation_folder is not None:
            ImageLabelClassification(
                self.validation_folder_images_labels,
                self.label2class_mapping,
                self.validation_folder_invalid_images_labels,
                self.folder_path,
                "Validation Folder Labels Summary ",
                "validation_folder_labels_summary",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "validation_folder_labels_summary"
            )
        if self.split_training:
            ImageLabelClassification(
                self.training_labels,
                self.label2class_mapping,
                None,
                self.folder_path,
                "Training Data After Splitting Labels Summary ",
                "training_after_splitting_labels_summary",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "training_after_splitting_labels_summary"
            )
            ImageLabelClassification(
                self.validation_labels,
                self.label2class_mapping,
                None,
                self.folder_path,
                "Validation Data After Splitting Labels Summary ",
                "validation_after_splitting_labels_summary",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "validation_after_splitting_labels_summary"
            )

    def make_garphs_sample(self):
        DisplaySampleImagesClassification(
            self.training_folder_images_paths,
            self.training_folder_images_labels,
            self.label2class_mapping,
            self.folder_path,
            title=" Random Samples of Training Folder",
            file_name="training_folder_random_samples",
        ).build_graph()
        self.report_data["graphs"]["images_name"].append(
            "training_folder_random_samples"
        )
        if self.validation_folder is not None:
            DisplaySampleImagesClassification(
                self.validation_folder_images_paths,
                self.validation_folder_images_labels,
                self.label2class_mapping,
                self.folder_path,
                title="Random Samples of Validation Folder",
                file_name="validation_folder_random_samples",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "validation_folder_random_samples"
            )
        if self.split_training:
            DisplaySampleImagesClassification(
                self.training_paths,
                self.training_labels,
                self.label2class_mapping,
                self.folder_path,
                title="Samples of Training Data After Splitting",
                file_name="training_after_splitting_random_samples",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "training_after_splitting_random_samples"
            )
            DisplaySampleImagesClassification(
                self.validation_paths,
                self.validation_labels,
                self.label2class_mapping,
                self.folder_path,
                title="Samples of Validation Data After Splitting",
                file_name="validation_after_splitting_random_samples",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "validation_after_splitting_random_samples"
            )

    def make_graphs_loader(self):
        training_images, training_labels = next(iter(self.training_loader))
        n_samples = min(5, len(training_images))

        training_images, training_labels = (
            training_images[:n_samples],
            training_labels[:n_samples],
        )
        ClassificationImagesAfterLoader(
            training_images,
            training_labels,
            self.label2class_mapping,
            self.folder_path,
            title="Samples of Training Data After Data Loader",
            file_name="training_after_data_loader_samples",
        ).build_graph()
        self.report_data["graphs"]["images_name"].append(
            "training_after_data_loader_samples"
        )
        if self.validation_loader is not None:
            validation_images, validation_labels = next(iter(self.validation_loader))
            n_samples = min(5, len(validation_images))
            validation_images, validation_labels = (
                validation_images[:n_samples],
                validation_labels[:n_samples],
            )
            ClassificationImagesAfterLoader(
                validation_images,
                validation_labels,
                self.label2class_mapping,
                self.folder_path,
                title="Samples of Validation Data After Data Loader",
                file_name="validation_after_data_loader_samples",
            ).build_graph()
            self.report_data["graphs"]["images_name"].append(
                "validation_after_data_loader_samples"
            )

    def make_graphs(self):
        self.report_data["graphs"] = {
            "folder_path": self.folder_path,
            "images_name": [],
        }
        self.make_graphs_label_summary()
        self.make_garphs_sample()
        self.make_graphs_loader()

    def get_loaders(self):
        training_dataset = ClassificationImageDataset(
            self.training_paths,
            self.training_labels,
            self.image_size,
            self.augment,
            self.augmentation_probability,
            self.horizontal_flip,
            self.vertical_flip,
            self.rotation,
            self.rotation_angle,
            self.brightness,
            self.brightness_factors,
            self.contrast,
            self.contrast_factors,
        )
        self.training_loader = DataLoader(
            training_dataset, batch_size=self.batch_size, shuffle=True
        )
        self.validation_loader = None
        if self.validation_paths is not None:
            validation_dataset = ClassificationImageDataset(
                self.validation_paths,
                self.validation_labels,
                self.image_size,
                False,
                self.augmentation_probability,
                self.horizontal_flip,
                self.vertical_flip,
                self.rotation,
                self.rotation_angle,
                self.brightness,
                self.brightness_factors,
                self.contrast,
                self.contrast_factors,
            )
            self.validation_loader = DataLoader(
                validation_dataset, batch_size=self.batch_size, shuffle=False
            )

    def common_preprocessing(self):
        self.get_classes_mapping()
        (
            self.training_folder_images_paths,
            self.training_folder_images_labels,
        ) = self.data_preparing(
            self.training_folder,
            self.training_folder_invalid_images,
            self.training_folder_invalid_images_labels,
            self.training_class,
        )

        if self.validation_folder is not None:
            (
                self.validation_folder_images_paths,
                self.validation_folder_images_labels,
            ) = self.data_preparing(
                self.validation_folder_images,
                self.validation_folder_invalid_images,
                self.validation_folder_invalid_images_labels,
                self.validation_class,
            )
        self.data_overview()
        self.splitting(
            self.training_folder_images_paths,
            self.validation_folder_images_paths,
            self.training_folder_images_labels,
            self.validation_folder_images_labels,
        )
        self.get_loaders()
        self.make_graphs()

        report = ImagePreprocessingReport(self.folder_path, self.report_data)
        report.execute()
        return self.training_loader, self.validation_loader



import os
import shutil
import tempfile

import numpy as np
import pandas as pd
import pytest
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import StandardScaler

from accelera.src.automl.core.classical_training_preprocessing import (
    ClassicalTrainingPreprocessing,
)
from accelera.src.automl.wrappers.frequency_encoder_transform import (
    FrequencyEncoderTransform,
)
from accelera.src.automl.wrappers.IQR_transform import IQRTransform
from accelera.src.utils.preprocessing import check_path_exists
from accelera.src.utils.preprocessing import load_pickle


class TestClassicalTrainingPreprocessing:
    @pytest.fixture(autouse=True)
    def temp_folder(self):
        self.temp_dir = tempfile.mkdtemp()
        yield self.temp_dir
        shutil.rmtree(self.temp_dir)

    @pytest.fixture(autouse=True)
    def sample_data(self):
        self.df_classification = pd.DataFrame(
            {
                "ID": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16],
                "const_feature": [
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                    5,
                ],
                "most_nulls_feature": [
                    1,
                    2,
                    None,
                    None,
                    3,
                    None,
                    None,
                    4,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ],
                "binary_feature": [
                    0,
                    1,
                    0,
                    1,
                    0,
                    1,
                    1,
                    0,
                    1,
                    0,
                    1,
                    0,
                    1,
                    0,
                    1,
                    0,
                ],
                "ordinal_feature": [
                    1,
                    2,
                    3,
                    4,
                    1,
                    2,
                    3,
                    4,
                    1,
                    2,
                    3,
                    4,
                    4,
                    3,
                    2,
                    1,
                ],
                "one_hot_feature": [
                    "A",
                    "B",
                    "C",
                    "A",
                    "B",
                    "C",
                    "A",
                    "B",
                    "C",
                    "A",
                    "B",
                    "C",
                    "A",
                    "B",
                    "C",
                    "A",
                ],
                "continuous_feature": [
                    10.5,
                    20.3,
                    30.1,
                    40.2,
                    50.5,
                    60.6,
                    70.7,
                    80.8,
                    90.9,
                    100.1,
                    110.2,
                    120.3,
                    130.4,
                    140.5,
                    150.6,
                    160.7,
                ],
                "frequency_feature": [
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                    "J",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                    "G",
                    "H",
                ],
                "text_feature": [
                    "This is a sample text.",
                    "Another example of text data.",
                    "Text data for testing.",
                    "Sample text for NLP tasks.",
                    "More text data here.",
                    "Text data for classification.",
                    "NLP text sample.",
                    "Text data example.",
                    "Sample text for analysis.",
                    "Another text data point.",
                    "Text data for model training.",
                    "Text data for evaluation.",
                    "Sample text input.",
                    "Text data for processing.",
                    "Another sample text.",
                    "Final text data point.",
                ],
                "Name_feature": [
                    "Alice",
                    "Bob",
                    "Charlie",
                    "David",
                    "Eve",
                    "Frank",
                    "Grace",
                    "Heidi",
                    "Ivan kan ",
                    "Judy Karl Karl",
                    "Karl Karl Karl",
                    "Liam Karl",
                    "Mia Karl",
                    "Nina",
                    "Oscar",
                    "Peggy",
                ],
                "target": [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
            }
        )
        self.df_classification = pd.concat(
            [self.df_classification, self.df_classification.iloc[[0]]],
            ignore_index=True,
        )

    def test_classical_training_preprocessing_setup(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        assert training_preprocessing.df is not None
        assert training_preprocessing.target_col == "target"
        assert training_preprocessing.problem_type == "classification"
        assert check_path_exists(self.temp_dir, "")

    def test_classical_errors_check(self):
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=self.df_classification,
                target_col="col1",
                problem_type="classification",
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=self.df_classification,
                target_col="target",
                problem_type="Crassification",
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=None,
                target_col="target",
                problem_type="regression",
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=self.df_classification,
                target_col="target",
                problem_type="regression",
                folder_path=self.temp_dir,
            )
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=self.df_classification,
                target_col="target",
                problem_type="regression",
                folder_path=None,
            )
        with pytest.raises(ValueError):
            ClassicalTrainingPreprocessing(
                df=self.df_classification,
                target_col="target",
                problem_type=None,
                folder_path=self.temp_dir,
            )

    def test_classical_data_overview(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        training_preprocessing.data_overview()
        assert training_preprocessing.report_data is not None
        assert "data_overview" in training_preprocessing.report_data
        data_overview = training_preprocessing.report_data["data_overview"]
        assert data_overview["shape"] == (17, 11)
        assert data_overview["duplicates_sum"] == 1
        assert data_overview["duplicates_percentage"] == float(1 / 17) * 100
        assert data_overview["numerical_describe"] is not None
        assert data_overview["categorical_describe"] is not None
        assert data_overview["data_head"].shape == (5, 11)
        assert data_overview["lower_data_head"].shape == (5, 11)
        assert data_overview["missing_values"].equals(
            self.df_classification.isnull().sum()
        )

    def test_classical_drop_dupplicates(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        shape_before_drop = training_preprocessing.df.shape
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        shape_after_drop = training_preprocessing.df.shape
        assert shape_after_drop[0] == shape_before_drop[0] - 1
        assert shape_after_drop[1] == shape_before_drop[1]
        assert (
            training_preprocessing.report_data["drop_duplicates"]["shape"]
            == shape_after_drop
        )
        assert self.df_classification.duplicated().sum() == 0
        assert (
            training_preprocessing.report_data["drop_duplicates"]["duplicates_sum"]
            == 0
        )
        assert (
            training_preprocessing.report_data["drop_duplicates"][
                "duplicates_percentage"
            ]
            == 0
        )

    def test_classical_split_data(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train_1, X_val_1, y_train_1, y_val_1 = train_test_split(
            training_preprocessing.df.drop(columns=["target"]),
            training_preprocessing.df["target"],
            test_size=0.2,
            random_state=42,
        )
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        assert X_train.shape[0] == X_train_1.shape[0]
        assert X_val.shape[0] == X_val_1.shape[0]
        assert y_train.shape[0] == y_train_1.shape[0]
        assert y_val.shape[0] == y_val_1.shape[0]
        assert (
            training_preprocessing.report_data["split"]["X_train_shape"]
            == X_train.shape
        )
        assert (
            training_preprocessing.report_data["split"]["X_val_shape"] == X_val.shape
        )
        assert (
            training_preprocessing.report_data["split"]["y_train_shape"]
            == y_train.shape
        )
        assert (
            training_preprocessing.report_data["split"]["y_val_shape"] == y_val.shape
        )
        assert X_train.equals(X_train_1)
        assert X_val.equals(X_val_1)
        assert y_train.equals(y_train_1)
        assert y_val.equals(y_val_1)

    def test_classical_get_info(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        assert len(col_drop) == 5
        assert "ID" in col_drop
        assert "const_feature" in col_drop
        assert "most_nulls_feature" in col_drop
        assert "text_feature" in col_drop
        assert col_drop["ID"] == "It is above unique_threshold 0.9"
        assert col_drop["const_feature"] == "The column is constant"
        assert (
            col_drop["most_nulls_feature"] == "Missing above missing_threshold 0.5"
        )
        assert col_drop["Name_feature"] == "It is above unique_threshold 0.9"
        assert col_drop["text_feature"] == "It is above unique_threshold 0.9"
        assert isinstance(info, dict)

    def test_classical_drop_columns(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        for col in col_drop.keys():
            assert col not in X_train.columns
            assert col not in X_val.columns
        assert X_train.shape[1] == 5
        assert X_val.shape[1] == 5
        assert check_path_exists(self.temp_dir, "col_drop.pkl")
        loaded_col_drop = load_pickle(self.temp_dir, "col_drop.pkl")
        assert loaded_col_drop == col_drop

    def test_classical_detect_column_types(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        assert set(binary_cols) == {"binary_feature"}
        assert set(one_hot_cols) == {"one_hot_feature"}
        assert set(frequency_cols) == {
            "frequency_feature",
        }
        assert set(numerical_cols) == {"continuous_feature"}
        assert set(ordinal_cols) == {"ordinal_feature"}

    def test_classical_make_graphs(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        training_preprocessing.make_graphs(X_train, y_train, info)
        graphs_path = os.path.join(self.temp_dir, "graphs")
        assert check_path_exists(graphs_path, "")
        graphs_files = os.listdir(graphs_path)
        expected_num_graphs = (
            len(binary_cols)
            + len(numerical_cols)
            + len(one_hot_cols)
            + len(frequency_cols)
            + len(ordinal_cols)
            + 2
        )
        assert len(graphs_files) == expected_num_graphs
        assert "binary_feature.png" in graphs_files
        assert "continuous_feature.png" in graphs_files
        assert "one_hot_feature.png" in graphs_files
        assert "frequency_feature.png" in graphs_files
        assert "ordinal_feature.png" in graphs_files
        assert "target.png" in graphs_files
        assert "correlation_matrix.png" in graphs_files

    def test_classical_features_preprocessing(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        assert X_train_processed.shape[0] == X_train.shape[0]
        assert X_val_processed.shape[0] == X_val.shape[0]
        assert check_path_exists(self.temp_dir, "training_preprocessor.pkl")
        assert check_path_exists(self.temp_dir, "feature_names.pkl")

    def test_classical_numerical_preprocessing_pipeline(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification[["continuous_feature", "target"]].copy(),
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        imputer = SimpleImputer(strategy="median")
        iqr_trasnform = IQRTransform(info, ["continuous_feature"])
        scaler = StandardScaler()
        X_train_manual = X_train[["continuous_feature"]].values
        X_val_manual = X_val[["continuous_feature"]].values
        X_train_imputed = imputer.fit_transform(X_train_manual)
        X_val_imputed = imputer.transform(X_val_manual)
        iqr_trasnform.fit(X_train_imputed)
        X_train_iqr = iqr_trasnform.transform(X_train_imputed)
        X_val_iqr = iqr_trasnform.transform(X_val_imputed)
        X_train_scaled = scaler.fit_transform(X_train_iqr)
        X_val_scaled = scaler.transform(X_val_iqr)
        assert X_train_processed.shape == X_train_scaled.shape
        assert X_val_processed.shape == X_val_scaled.shape
        assert np.allclose(X_train_processed, X_train_scaled, atol=1e-6)
        assert np.allclose(X_val_processed, X_val_scaled, atol=1e-6)

    def test_classical_binary_preprocessing_pipeline(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification[["binary_feature", "target"]].copy(),
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        imputer = SimpleImputer(strategy="most_frequent")
        ordinal_encoder = OrdinalEncoder(
            handle_unknown="use_encoded_value", unknown_value=-1
        )
        X_train_manual = X_train[["binary_feature"]].values
        X_val_manual = X_val[["binary_feature"]].values
        X_train_imputed = imputer.fit_transform(X_train_manual)
        X_val_imputed = imputer.transform(X_val_manual)
        X_train_encoded = ordinal_encoder.fit_transform(X_train_imputed)
        X_val_encoded = ordinal_encoder.transform(X_val_imputed)
        assert X_train_processed.shape == X_train_encoded.shape
        assert X_val_processed.shape == X_val_encoded.shape
        assert np.allclose(X_train_processed, X_train_encoded, atol=1e-6)
        assert np.allclose(X_val_processed, X_val_encoded, atol=1e-6)

    def test_classical_one_hot_preprocessing_pipeline(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification[["one_hot_feature", "target"]].copy(),
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        imputer = SimpleImputer(strategy="most_frequent")
        one_hot_encoder = OneHotEncoder(
            handle_unknown="ignore",
            sparse_output=False,
            drop="first",
        )
        X_train_manual = X_train[["one_hot_feature"]].values
        X_val_manual = X_val[["one_hot_feature"]].values
        X_train_imputed = imputer.fit_transform(X_train_manual)
        X_val_imputed = imputer.transform(X_val_manual)
        X_train_onehot = one_hot_encoder.fit_transform(X_train_imputed)
        X_val_onehot = one_hot_encoder.transform(X_val_imputed)
        assert X_train_processed.shape == X_train_onehot.shape
        assert X_val_processed.shape == X_val_onehot.shape
        assert np.allclose(X_train_processed, X_train_onehot, atol=1e-6)
        assert np.allclose(X_val_processed, X_val_onehot, atol=1e-6)
        assert X_train_processed.shape[1] == X_train["one_hot_feature"].nunique() - 1
        assert X_val_processed.shape[1] == X_train["one_hot_feature"].nunique() - 1

    def test_classical_frequency_preprocessing_pipeline(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification[["frequency_feature", "target"]].copy(),
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        imputer = SimpleImputer(strategy="most_frequent")
        frequency_encoder = FrequencyEncoderTransform()
        X_train_manual = X_train[["frequency_feature"]].values
        X_val_manual = X_val[["frequency_feature"]].values
        X_train_imputed = imputer.fit_transform(X_train_manual)
        X_val_imputed = imputer.transform(X_val_manual)
        X_train_processed_manual = frequency_encoder.fit_transform(X_train_imputed)
        X_val_processed_manual = frequency_encoder.transform(X_val_imputed)
        assert np.allclose(
            X_train_processed[:, 0], X_train_processed_manual.ravel(), atol=1e-6
        )
        assert np.allclose(
            X_val_processed[:, 0], X_val_processed_manual.ravel(), atol=1e-6
        )

    def test_classical_target_classification_preprocessing(self):
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=self.df_classification,
            target_col="target",
            problem_type="classification",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        y_train_processed, y_val_processed = (
            training_preprocessing.target_preprocessing(y_train, y_val, info)
        )
        mode = y_train.mode()[0]
        label_encoder = LabelEncoder()
        y_train_filled = y_train.fillna(mode)
        y_val_filled = y_val.fillna(mode)
        y_train_encoded = label_encoder.fit_transform(y_train_filled)
        y_val_encoded = label_encoder.transform(y_val_filled)
        assert y_train_processed.shape == y_train_encoded.shape
        assert y_val_processed.shape == y_val_encoded.shape
        assert np.allclose(y_train_processed, y_train_encoded, atol=1e-6)
        assert np.allclose(y_val_processed, y_val_encoded, atol=1e-6)
        assert check_path_exists(self.temp_dir, "target_preprocessor.pkl")
        assert check_path_exists(self.temp_dir, "target_info.pkl")

    def test_classical_target_regression_preprocessing(self):
        df_regression = self.df_classification.copy()
        df_regression["target"] = df_regression[
            "continuous_feature"
        ] + np.random.randn(len(df_regression))
        training_preprocessing = ClassicalTrainingPreprocessing(
            df=df_regression,
            target_col="target",
            problem_type="regression",
            folder_path=self.temp_dir,
            cardinality_threshold=6,
            max_unique_ordinal=8,
        )
        training_preprocessing.data_overview()
        training_preprocessing.drop_duplicates()
        X_train, X_val, y_train, y_val = training_preprocessing.split_data()
        info, col_drop = training_preprocessing.get_data_info(X_train, y_train)
        training_preprocessing.drop_col(X_train, X_val, col_drop)
        (
            binary_cols,
            numerical_cols,
            one_hot_cols,
            frequency_cols,
            ordinal_cols,
            _,
        ) = training_preprocessing.detect_column_types(X_train, info)
        X_train_processed, X_val_processed = (
            training_preprocessing.features_preprocessing(
                X_train,
                X_val,
                info,
                binary_cols,
                numerical_cols,
                one_hot_cols,
                frequency_cols,
                ordinal_cols,
            )
        )
        y_train_processed, y_val_processed = (
            training_preprocessing.target_preprocessing(y_train, y_val, info)
        )
        median = y_train.median()
        y_train_filled = y_train.fillna(median)
        y_val_filled = y_val.fillna(median)
        stander = StandardScaler()
        y_train_scaled = stander.fit_transform(
            y_train_filled.values.reshape(-1, 1)
        ).ravel()
        y_val_scaled = stander.transform(y_val_filled.values.reshape(-1, 1)).ravel()
        assert y_train_processed.shape == y_train_scaled.shape
        assert y_val_processed.shape == y_val_scaled.shape
        assert np.allclose(y_train_processed, y_train_scaled, atol=1e-6)
        assert np.allclose(y_val_processed, y_val_scaled, atol=1e-6)
        assert check_path_exists(self.temp_dir, "target_preprocessor.pkl")
        assert check_path_exists(self.temp_dir, "target_info.pkl")



import numpy as np

from accelera.src.accelera_pipe.core.report_base import ReportBase
from accelera.src.accelera_pipe.wrappers.display_array_multi import DisplayMultiArray
from accelera.src.accelera_pipe.wrappers.display_array_single import (
    DisplayArraySingle,
)
from accelera.src.accelera_pipe.wrappers.display_dict import DisplayDict
from accelera.src.accelera_pipe.wrappers.display_figure import DisplayFigure
from accelera.src.accelera_pipe.wrappers.display_single_number import (
    DisplaySingleNumber,
)
from accelera.src.accelera_pipe.wrappers.display_string import DisplayString
from accelera.src.accelera_pipe.wrappers.display_tuple_not_curve import (
    DisplayTupleNotCurve,
)
from accelera.src.utils.accelera_utils import create_folder


class GraphPipelineReport(ReportBase):
    def __init__(self, folderpath, results):
        super().__init__(folderpath)
        self.results = results
        create_folder(folderpath)
        self.metric_ids = []

    def handle_metric(self):
        final_metric = {}
        for i in range(len(self.metric_ids)):
            metric_name = self.results[i]["metric name"]
            metric_value = self.results[i]["result"]
            metric_plot_func = self.results[i]["plot_func"]
            metric_lables_name = self.results[i]["labels_name"]
            metric_headers_name = self.results[i]["headers_name"]
            metric_obj = {
                "metric id": self.metric_ids[i],
                "result": metric_value,
                "plot_func": metric_plot_func,
                "labels_name": metric_lables_name,
                "headers_name": metric_headers_name,
            }
            if metric_name in final_metric:
                final_metric[metric_name].append(metric_obj)
            else:
                final_metric[metric_name] = [metric_obj]
        return final_metric

    def metric_display(self):
        metric = self.handle_metric()
        metric_content = (
            "<h2> Metrics Summary</h2>\n"
            "<ul>\n"
            "<li>Each metric is displayed with its user-defined name,"
            " unique identifier (ID) and the corresponding results.</li>\n"
            "<li>Depending on the metric type, the results may include "
            "scalar values, arrays, dictionaries, "
            "strings, curves, or tuples.</li>\n"
            "<li>All metrics are presented in a structured "
            "and consistent format "
            "to facilitate clear interpretation and comparison.</li>\n"
            "</ul>\n"
        )
        for metric_name, values in metric.items():
            if isinstance(values[0]["result"], (int, float)):
                obj = DisplaySingleNumber(metric_name, values, self.folderpath)
                content = obj.execute()
            elif (
                isinstance(values[0]["result"], (np.ndarray))
                and values[0]["result"].ndim > 1
            ):
                if values[0]["plot_func"] is None:
                    obj = DisplayMultiArray(metric_name, values)
                    content = obj.execute()
                else:
                    obj = DisplayFigure(metric_name, values, self.folderpath)
                    content = obj.execute()
            elif (
                isinstance(values[0]["result"], (np.ndarray))
                and values[0]["result"].ndim == 1
            ):
                obj = DisplayArraySingle(metric_name, values)
                content = obj.execute()
            elif isinstance(values[0]["result"], dict):
                obj = DisplayDict(metric_name, values)
                content = obj.execute()
            elif isinstance(values[0]["result"], str):
                obj = DisplayString(metric_name, values)
                content = obj.execute()
            elif isinstance(values[0]["result"], (tuple)):
                if values[0]["plot_func"] is None:
                    obj = DisplayTupleNotCurve(metric_name, values, self.folderpath)
                    content = obj.execute()
                else:
                    obj = DisplayFigure(metric_name, values, self.folderpath)
                    content = obj.execute()
            metric_content = metric_content + "\n" + content
        return metric_content

    def execute(self):
        pass



import os


class ReportBase:
    def __init__(self, folderpath):
        self.folderpath = folderpath
        self.start_content = """
        <!DOCTYPE html>
        <html>
        <head>
          <meta charset="UTF-8">
          <meta name="viewport" content="width=device-width, initial-scale=1.0">
          <title>Accelera Report</title>
          <style>
          body {
            font-family: Arial, sans-serif;
            line-height: 1.6;
            background-color: #000000;
            color: #ffffff;
            margin-left:30px
            }
            h1 {
            color: #ffcc00;
            text-align: center;
            }
            h2{
            font-style:italic;
            text-align: center;
            }
            h3{
            color: #ffee00;
            }
            h4 {
            color: #ffcc00; 
            }
            
            table{
            border-collapse: collapse;
            width: 80%;
            }
             th, td {
            border: 1px solid #444;
            padding: 6px;
            text-align: center;
            }
            .metric-container {
            display: grid;
            grid-template-columns: repeat(2, minmax(0, 1fr));
            gap: 20px;
            }
            div{
                margin:40px 10px ;
            }
        </style>
        </head>
        <body>
        <h1>Accelera Report</h1>
        """
        self.end_content = """
        </body>
        </html>
        """

    def show_graphs(self, graphs):
        if graphs is None:
            return
        self.content += "<div>\n"
        self.content += "<h2>Graphs</h2>\n"
        folder_path = os.path.join(graphs["folder_path"], "graphs")
        if not os.path.exists(folder_path):
            self.content += (
                f"<h3>There is no any image in the folder {folder_path}</h3>\n"
            )
        else:
            for image_name in graphs["images_name"]:
                image_file = os.path.join(folder_path, f"{image_name}.png")
                if os.path.exists(image_file):
                    image_file = os.path.join(".", "graphs", f"{image_name}.png")
                    self.content += f"<img src='{image_file}' "
                    self.content += "style='max-width:100%; margin:10px 0;'/>\n"
        self.content += "</div>\n"

    def create_html_file(self, content):
        readme_path = os.path.join(self.folderpath, "report.html")
        with open(readme_path, encoding="utf-8", mode="w") as f:
            f.write(content)

    def execute(self):
        raise NotImplementedError("Subclasses must implement this method")



import pandas as pd

from accelera.src.accelera_pipe.core.report_base import ReportBase


class ImagePreprocessingReport(ReportBase):
    def __init__(self, folderpath, report_data):
        super().__init__(folderpath)
        self.content = ""
        self.report_data = report_data
        self.data_overview_training = self.report_data["data_overview"].get(
            "training_folder"
        )
        self.data_overview_validation = self.report_data["data_overview"].get(
            "validation_folder"
        )
        self.split_data = self.report_data.get("split_data")
        self.graphs = self.report_data.get("graphs")

    def show_unoreder_list(self, classes, title=None):
        if classes is not None:
            self.content += "<div>\n"
            if title is not None:
                self.content += f"{title}\n"
            self.content += "<ul>\n"
            for class_name in classes:
                self.content += f"<li>{class_name}</li>\n"
            self.content += "</ul>\n"
            self.content += "</div>\n"

    def show_paragraph_section(self, text, title=None):
        self.content += "<div>\n"
        if title is not None:
            self.content += f"{title}\n"
        self.content += f"<p>{text}</p>\n"
        self.content += "</div>\n"

    def show_table_section_from_df(self, dictionary, title=None):
        self.content += "<div>\n"
        if title is not None:
            self.content += f"{title}\n"
        self.content += dictionary.to_html(index=False)
        self.content += "</div>\n"

    def show_folder_overview(self, over_view, folder_type):
        self.content += "<div>\n"
        self.content += f"<h3>{folder_type} Folder Overview</h2>\n"
        self.show_unoreder_list(over_view.get("classes"), "<h4>Classes</h4>")
        if over_view.get("mapping") is not None:
            df = pd.DataFrame(
                over_view["mapping"].items(), columns=["Class", "Label"]
            )
            self.show_table_section_from_df(df, "<h4>Classes 2 Labels Mapping </h4>")
        self.show_paragraph_section(
            f"Count : {over_view['images_len']}", "<h4>Total Images </h4>"
        )
        self.content += "<div>\n"
        self.content += "<h4>Invalid Images</h4>\n"
        self.show_paragraph_section(f"Count : {over_view['invalid_len']}")
        self.show_unoreder_list(over_view["invalid_images"])
        self.content += "</div>\n"
        self.show_table_section_from_df(
            over_view["random_sample"], "<h4>Random Sample</h4>\n"
        )
        self.content += "</div>\n"

    def show_overview(self):
        self.content += "<div>\n"
        self.content += "<h2>Data Overview</h2>\n"
        self.show_folder_overview(self.data_overview_training, "Training")
        if self.data_overview_validation is not None:
            self.show_folder_overview(self.data_overview_training, "Validation")
        self.content += "</div>\n"

    def show_data_split(self):
        if self.split_data is None:
            return
        self.content += "<div>\n"
        self.content += "<h2>Splitting</h2>\n"
        self.show_paragraph_section(
            f"{self.split_data['validation_size']}",
            "<h3>Validation Size</h3>\n",
        )
        self.content += "<div>\n"
        self.content += "<h3>Training After Splitting</h3>\n"
        self.show_paragraph_section(
            f"{self.split_data['training_data_size']}", "<h4>Total Images </h4>"
        )
        self.show_table_section_from_df(
            self.split_data["random_training_sample"],
            "<h4>Random Sample</h4>\n",
        )

        self.content += "</div>\n"
        self.content += "<div>\n"
        self.content += "<h3>Validation After Splitting</h3>\n"
        self.show_paragraph_section(
            f"{self.split_data['validation_data_size']}",
            "<h4>Total Images </h4>",
        )
        self.show_table_section_from_df(
            self.split_data["random_validation_sample"],
            "<h4>Random Sample</h4>\n",
        )
        self.content += "</div>\n"
        self.content += "</div>\n"

    def execute(self):
        self.show_overview()
        self.show_data_split()
        self.show_graphs(self.graphs)
        full_content = self.start_content + self.content + self.end_content
        self.create_html_file(full_content)



import pandas as pd

from accelera.src.custom.transformer import CustomTransformer


class DateFeatureExtractor(CustomTransformer):
    def __init__(self, cols):
        self.cols = cols
        self.feature_names_ = []
        self.median_ = {}

    def fit(self, X, y=None):
        self.feature_names_ = []
        self.median_ = {}

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X, columns=self.cols)
        for col in self.cols:
            self.feature_names_.extend(
                [
                    f"{col}_year",
                    f"{col}_month",
                    f"{col}_day",
                    f"{col}_weekday",
                    f"{col}_hour",
                ]
            )
            self.median_[col] = X[col].median()
        return self

    def transform(self, X):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X, columns=self.cols)
        X_output = pd.DataFrame()
        for col in self.cols:
            col_data = pd.to_datetime(X[col], errors="coerce")
            col_date = col_data.fillna(self.median_[col])
            X_output[f"{col}_year"] = col_date.dt.year.fillna(0).astype(int)
            X_output[f"{col}_month"] = col_date.dt.month.fillna(0).astype(int)
            X_output[f"{col}_day"] = col_date.dt.day.fillna(0).astype(int)
            X_output[f"{col}_weekday"] = col_date.dt.weekday.fillna(0).astype(int)
            X_output[f"{col}_hour"] = col_date.dt.hour.fillna(0).astype(int)
        return X_output.values

    def get_feature_names_out(self, input_features=None):
        return self.feature_names_



import os

from accelera.src.accelera_pipe.core.pipeline_base import PipelineBase
from accelera.src.utils.accelera_utils import serialize


class ExecutedGraph(PipelineBase):
    def __init__(self, executed_graph):
        super().__init__(_graph=executed_graph)

    def __call__(self, X, y_true=None):
        if y_true is not None:
            self._PipelineBase__graph.enableDisableMetrics(
                y_true=y_true, enable=True
            )

        results = self._PipelineBase__graph.execute(X)

        if y_true is not None:
            self._PipelineBase__graph.enableDisableMetrics(enable=False)

        return results

    def save(self, directory):
        if not os.path.exists(directory):
            os.makedirs(directory)
        serialize(self, os.path.join(directory, "pipeline.xml"))
        # TODO: we need also to save models and other
        # data needed later for loading
        return self
